from mammut.curriculum.classroom.ray_classroom import RayClassroom
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
from typing import Set
import ray
import mammut.curriculum.report as report


class ClassroomPool:
    """Pool of RayClassroom instances for client classes.

    RayClassrooms are instantiated at initialization time.
    The number of classrooms is fixed.
    """

    def __init__(
        self,
        classrooms: int,
        package_id: str,
        standard_spreadsheet_id: str,
        mammut_session_context: MammutSessionContext,
        enable_ktt_interpretation: bool,
        package_id_is_spreadsheet: bool
    ):
        """

        Args:
            classrooms(int): number of classrooms to hold in this class.
            package_id(str): package spreadsheet ID / presentation ID to forward to remote classrooms.
            standard_spreadsheet_id(str): standard spreadsheet id to forward to remote classrooms.
            mammut_session_context: The session context to pass down to remote classrooms processes.
            enable_ktt_interpretation (bool): if True, remote classrooms will load an instance of Ktt interpretation
                 course
             package_id_is_spreadsheet (bool): True if the package ID is from a presentation.
        """

        self._mammut_session_context: MammutSessionContext = mammut_session_context

        report.send_message(
            report.CurriculumGeneralDebugMessage(
                "Creating classrooms in remote workers"
            )
        )

        self._available: Set[RayClassroom] = set(
            [
                RayClassroom.remote(self._mammut_session_context)
                for i in range(classrooms)
            ]
        )

        self._known_classrooms = set(
            [ray.get(classroom.id.remote()) for classroom in self._available]
        )

        [classroom.initialize.remote() for classroom in self._available]

        [
            classroom.add_curriculum.remote(
                package_id,
                standard_spreadsheet_id,
                enable_ktt_interpretation=enable_ktt_interpretation,
                package_id_is_spreadsheet=package_id_is_spreadsheet
            )
            for classroom in self._available
        ]

    def available(self) -> bool:
        """Returns True if there are more classrooms available.
        """
        return bool(self._available)

    def take_out(self):
        """Take out a classroom from this pool.

        After calling this method, there will be one less available
        course.

        Returns:
            A classroom instance. None if there aren't any available.
        """
        if self.available():
            report.send_message(
                report.CurriculumGeneralDebugMessage(
                    "Classroom take_out from classrooms pool."
                )
            )
            return self._available.pop()
        else:
            return None

    def put_in(self, classroom: RayClassroom):
        """Put the element back in the pool.

        Raises:
            ValueError: when an unknown classrooms tries to be inserted in the pool.
        """
        if ray.get(classroom.id.remote()) not in self._known_classrooms:
            message = (
                "Wrong classroom argument. No previously known classroom."
            )
            report.send_message(report.CurriculumGeneralErrorMessage(message))
            raise ValueError(message)

        report.send_message(
            report.CurriculumGeneralDebugMessage(
                "Classroom put_in to classrooms pool."
            )
        )
        self._available.add(classroom)



import enum
import logging
from abc import ABC, abstractmethod
from typing import Dict
from mammut.curriculum.core.mammut_session_context import MammutSessionContext

log = logging.getLogger(__name__)


class ModelBase(ABC):
    """ Base API for Curriculum models.

    This class exposes a basic API that allows to use
    different models for Curriculum client code.

    """

    def __init__(self, parameters: Dict, course_id: int, lesson_id: int):
        """
        Args:
            parameters: Dictionary with lessons parameters to forward to this model.
            course_id (int): the course ID to which this model belongs.
            lesson_id (int): the lesson ID to which this model belongs.
        """
        self.parameters: Dict = parameters
        self.course_id: int = course_id
        self.lesson_id: int = lesson_id

    @abstractmethod
    def train(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Train the model this class implements.

        Args:
            **kwargs: Named parameters required for the implementation
                in the child class.
        """
        pass

    @abstractmethod
    def save(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Persist the model in memory.

        Args:
            mammut_session_context: Mammut session context with general information about
                current curriculum instance.
            **kwargs: Any other custom attribute.
        """
        pass

    def _get_model_file_system_path(self, file_system_prefix_path: str) -> str:
        """Return the file system path for saving this model.

        The prefix path is the same for all curriculum models, with
        specific directories for each course/lesson. This method concatenates
        those specific directories.

        An example of a full file system path for a curriculum model is:
            /home/user/.mammut-py/curriculum/14/3

            The prefix path is: /home/user/.mammut-py/curriculum/
            The last folders: /14/3 refers to the course 14 and lesson 3

        Args:
            file_system_prefix_path: prefix of the path.
        """
        save_folder = (
            file_system_prefix_path + f"/{self.course_id}/{self.lesson_id}"
        )
        return save_folder

    @classmethod
    def get_model_file_system_path(
        cls, file_system_prefix_path: str, course_id: str, lesson_id: str
    ) -> str:
        """Returns the file system path for a trained model, given a course ID
        and a lesson ID.

        Args:
            file_system_prefix_path: prefix of the path.
            course_id: some course ID
            lesson_id: some lesson ID.
        """
        save_folder = file_system_prefix_path + f"/{course_id}/{lesson_id}"
        return save_folder



# coding=utf-8
from mammut.common.simulator_utilities import *
from ipywidgets import *


def get_solve_totbl_box():
    form_values = dict()
    responseLayout = Layout(width="100%", height="50px")
    responseTextBox = widgets.Text(
        value="", placeholder="Respuesta", description="Nombre:", layout=responseLayout
    )
    form_values["responseTextBox"] = responseTextBox

    solveTotblButton = widgets.Button(
        description="Crear Nuevo ...", button_style="primary"
    )

    solveTotblBox = widgets.VBox([responseTextBox, solveTotblButton])
    return solveTotblBox, solveTotblButton, form_values



# coding=utf-8

from ipywidgets import *
from django.template.loader import get_template
from mammut.common.corpus.dictionary_corpus import DictionaryCorpus
from mammut.common.annotations import MammutEventAnnotation

functionals_template = get_template("functionals_help.html")


def annotations_widget_section(package):
    def restitution_section(event_annotation, single_annotator):
        # creamos el control donde se debe introducir la restutucion
        # restitution_str = ''
        restitution_input = widgets.Text(
            value="",
            placeholder="tu restitucion",
            description="Restitucion:",
            layout=Layout(width="250px"),
        )

        def on_restitution_text_value_change(change):
            return

        restitution_input.observe(on_restitution_text_value_change, names="value")

        delete_restitution_button = widgets.Button(
            description="eliminar restituciones",
            button_style="",
            tooltip="eliminar Restituciones",
            icon="",
            layout=Layout(width="200px"),
        )

        def handle_delete_restitution(sender):
            package.delete_restitutions(event_annotation)
            new_anotator = event_annotation.get_annotations_elements("exchange")[
                0
            ]  # TODO:revisar este cero
            restitution_input.value = ""
            mixed_box.children = [restitution_box, new_anotator]

        delete_restitution_button.on_click(handle_delete_restitution)

        save_restitution_button = widgets.Button(
            description="Guardar restitucion",
            button_style="",
            tooltip="Guardar Restitucion",
            icon="",
            layout=Layout(width="200px"),
        )

        def handle_save_restitution(sender):
            package.add_restitution(event_annotation, restitution_input)
            new_anotator = event_annotation.get_annotations_elements("exchange")[
                0
            ]  # TODO:revisar este cero
            restitution_input.value = ""
            mixed_box.children = [restitution_box, new_anotator]

        save_restitution_button.on_click(handle_save_restitution)

        restitution_box = widgets.HBox(
            [restitution_input, save_restitution_button, delete_restitution_button]
        )
        mixed_box = widgets.VBox([restitution_box, single_annotator])
        return mixed_box

    def dimension_selector(
        event_annotation,
    ):  # TODO:  poner cual dimension se esta editando??
        dimensions_dict = {}
        for dimension in event_annotation.dimensions:
            dimensions_dict[dimension] = dimension
        dimensions_dict["None"] = "None"

        dimensions_dropdown = widgets.Dropdown(
            options=dimensions_dict,
            value="None",
            description="select dimension:",
            layout=Layout(width="500px"),
        )
        save_annotation_button = widgets.Button(
            description="Guardar Anotacion",
            button_style="",
            tooltip="Guardar Anotacion",
            icon="",
            layout=Layout(width="200px"),
        )

        def handle_save_annotation(change):
            if (
                len(
                    list(
                        event_annotation.documents_by_dimension[
                            dimensions_dropdown.value
                        ].ann_obj.get_textbounds()
                    )
                )
                > 0
            ):
                event_annotation.add_annotation_to_sheet(dimensions_dropdown.value)
                event_annotation.is_annotated_dimention[
                    dimensions_dropdown.value
                ] = True
                if (
                    event_annotation.current_dimension(dimensions_dropdown.value)
                    is not None
                ):
                    dimensions_dropdown_change(
                        {
                            "new": event_annotation.current_dimension(
                                dimensions_dropdown.value
                            )
                        }
                    )
                    # TODO: se simula el objeto change, deberia crear un nuevo metodo en lugar de utilizar el mismo dimension_dropdow_change

        save_annotation_button.on_click(handle_save_annotation)

        def dimensions_dropdown_change(change):
            if change["new"] is not "None":
                annotation_elements = event_annotation.get_annotations_elements(
                    change["new"]
                )
                if (
                    event_annotation.current_dimension(dimensions_dropdown.value)
                    is not None
                ):
                    dimensions_dropdown.value = event_annotation.current_dimension(
                        dimensions_dropdown.value
                    )
                if len(annotation_elements) == 1:
                    if not isinstance(corpus_dropdown.value, DictionaryCorpus):
                        box = [
                            dimensions_dropdown,
                            restitution_section(
                                event_annotation, annotation_elements[0]
                            ),
                            save_annotation_button,
                        ]
                    else:
                        box = (
                            [dimensions_dropdown]
                            + annotation_elements
                            + [save_annotation_button]
                        )
                else:
                    box = (
                        [dimensions_dropdown]
                        + annotation_elements
                        + [save_annotation_button]
                    )
                dimension_selector_elements.children = box
            else:
                dimension_selector_elements.children = [dimensions_dropdown]

        dimensions_dropdown.observe(dimensions_dropdown_change, names="value")

        dimension_selector_elements = widgets.VBox([dimensions_dropdown])

        return dimension_selector_elements

    def event_selector(corpus):
        annotation_object = package.annotations
        all_events = annotation_object.get_event_messages(corpus)
        annotation_events = annotation_object.get_annotation_event_messages(corpus)
        all_events_dict = {}
        for annotaion_event_message in annotation_events:
            try:
                selector_key = corpus.hash_lemma_dict[
                    annotaion_event_message.event[
                        MammutEventAnnotation.SCENARIO_INDEX_ID
                    ]
                ]
            except AttributeError:
                selector_key = annotaion_event_message.event["original_text"]

            if annotaion_event_message.event["event_entry"] > 0:
                all_events_dict[
                    selector_key
                    + "-"
                    + str(annotaion_event_message.event["event_entry"])
                ] = annotaion_event_message
            else:
                all_events_dict[selector_key] = annotaion_event_message
        all_events_dict["None"] = "None"

        events_dropdown = widgets.Dropdown(
            options=all_events_dict,
            value="None",
            description="all events:",
            layout=Layout(width="800px"),
        )

        def events_dropdown_change(change):
            if change["new"] is not "None":
                anotation_widget.children = [
                    corpus_dropdown,
                    events_dropdown,
                    dimension_selector(change["new"]),
                ]
            else:
                anotation_widget.children = [corpus_dropdown, events_dropdown]

        events_dropdown.observe(events_dropdown_change, names="value")
        return events_dropdown

    all_corpus = package.corpus_map.get_all_corpus()
    all_corpus_dict = {}
    for i in all_corpus:
        if i.annotable:
            all_corpus_dict[i.sheet_title] = i
    all_corpus_dict["None"] = "None"
    corpus_dropdown = widgets.Dropdown(
        options=all_corpus_dict,
        value="None",
        description="select corpus:",
        layout=Layout(width="500px"),
    )

    def corpus_dropdown_change(change):
        if change["new"] != "None":
            anotation_widget.children = [corpus_dropdown, event_selector(change["new"])]
        else:
            anotation_widget.children = [corpus_dropdown]

    corpus_dropdown.observe(corpus_dropdown_change, names="value")

    anotation_widget = widgets.VBox([corpus_dropdown])
    return anotation_widget


def functionals_help_widget_section(package):
    def display_information(functional_lemma_descriptor):
        definitions = [
            {"index": i.index, "description": i.get_description()}
            for i in functional_lemma_descriptor.definitions
        ]
        context_dict = {"functionals_indexes": definitions}
        functionals_help_html = functionals_template.render(context_dict)
        functionals_help = widgets.HTML(value=functionals_help_html)
        return functionals_help

    def functionals_selector(option):
        lang, pos_id = option.split("-")
        all_functionals = package.standard.functionals  # debe ser un diccionario
        option_functionals = {}
        for key, val in all_functionals.items():
            if val.pos_paradigm == pos_id and val.regional_settings_lemma == lang:
                option_functionals[val.lemma] = val
        option_functionals["None"] = "None"

        functionals_dropdown = widgets.Dropdown(
            options=option_functionals,
            value="None",
            description="select " + option + ":",
            layout=Layout(width="600px"),
        )

        def functionals_dropdown_change(change):
            if change["new"] != "None":
                functionals_help_widget.children = [
                    functional_options_dropdown,
                    functionals_dropdown,
                    display_information(change["new"]),
                ]
            else:
                functionals_help_widget.children = [
                    functional_options_dropdown,
                    functionals_dropdown,
                ]

        functionals_dropdown.observe(functionals_dropdown_change, names="value")

        return functionals_dropdown

    functional_pos = []
    functional_options = {}
    for key, val in package.standard.functionals.items():
        if not val.is_bound:
            functional_pos.append(val.pos_paradigm)
    functional_pos = set(functional_pos)
    for functional_pos_id in functional_pos:
        pos_descriptor = package.standard.pos_descriptor_list[functional_pos_id]
        for lang in pos_descriptor.regional_settings_list:
            functional_options[f"{lang}-{pos_descriptor.name}"] = (
                lang + "-" + pos_descriptor.id
            )
    # TODO: si hago un reduce de las clases de todos los funcionales?sera el mismo widget para describir otros funcionales o solo para prep/conj
    functional_options["None"] = "None"
    functional_options_dropdown = widgets.Dropdown(
        options=functional_options,
        value="None",
        description="select pos pardigm:",
        layout=Layout(width="600px"),
    )

    def functionals_options_dropdown_change(change):
        if change["new"] != "None":
            functionals_help_widget.children = [
                functional_options_dropdown,
                functionals_selector(change["new"]),
            ]
        else:
            functionals_help_widget.children = [functional_options_dropdown]

    functional_options_dropdown.observe(
        functionals_options_dropdown_change, names="value"
    )

    functionals_help_widget = widgets.VBox([functional_options_dropdown])
    return functionals_help_widget



from abc import ABC
from typing import List
from mammut.curriculum.core.utils import LanguageLevel
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
import dataclasses


@dataclasses.dataclass
class CourseState(ABC):
    """Data class for courses.

    It contains all the relevant state for the course, without extra processing logic.

    A course is approved if the total sum of weights for the approved
    problems surpasses or equals the approval_criteria.

    If a course can't be approved, it will be reprocessed in the next
    academic period. If for some reason a course can't never be approved,
    the retries_count for the course will avoid an infinite loop of an academic period.

    Args:
        course_id(int): 'id' value in datasource retrieved data.
        academic_period(int): 'academic_period' value in datasource retrieved data.
        prerequisites(List[int]): 'prerequisites' value in datasource retrieved data.
        language_level: 'language_level' value in datasource retrieved data.
        course_name(str): 'course_name' value in datasource retrieved data.
        approval_criteria(float): approval normalized criteria for this course.
        retry(int): 'retry' string value to indicate how many times a course can be processed before
            stopping the curriculum.
        retries_count(int)
        regional_settings(str): 'regional_settings' value in datasource retrieved data.
        last_score(float): last registered score for this course.
        perfect_score(float): the perfect score of the course if all problems are passed.
        mammut_session_context: The curriculum session context where this course instance exists.
    """

    course_id: int
    academic_period: int
    prerequisites: List[int]
    language_level: LanguageLevel
    course_name: str
    _approval_criteria: float
    _retry_value: int
    _retries_count: int
    regional_settings: str
    _last_score: float
    _perfect_score: float
    mammut_session_context: MammutSessionContext

    def __hash__(self):
        return self.course_id

    def __eq__(self, other):
        return (
            isinstance(other, CourseState)
            and self.course_id == other.course_id
        )



# coding=utf-8
from datetime import datetime
from enum import Enum

def get_datetime_from_rfc3339(datetime_str: str):
    d = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%fZ')
    return d

class StringEnum(Enum):
    @classmethod
    def get_values(cls):
        return list(map(lambda e: e[1].value, cls.__members__.items()))

    @classmethod
    def contains_value(cls, value: str):
        return value in cls.get_values()


class RegionalSettingsType(StringEnum):
    ES = 'es'
    ES_VE = 'es-ve'
    EN = 'en'
    EN_US = 'en-us'


class YesNoType(StringEnum):
    Yes = 'Yes'
    No = 'No'

class DaysType(StringEnum):
    Mo = 'Mo'
    Tu = 'Tu'
    We = 'We'
    Th = 'Th'
    Fr = 'Fr'
    Sa = 'Sa'
    Su = 'Su'

DAYS_FOR_SP = {
    DaysType.Mo.value: 'Lunes',
    DaysType.Tu.value: 'Martes',
    DaysType.We.value: 'Miercoles',
    DaysType.Th.value: 'Jueves',
    DaysType.Fr.value: 'Viernes',
    DaysType.Sa.value: 'Sabado',
    DaysType.Su.value: 'Domingo'
}

DAYS_FOR_EN = {
    DaysType.Mo.value: 'Monday',
    DaysType.Tu.value: 'Tuesday',
    DaysType.We.value: 'Wednesday',
    DaysType.Th.value: 'Thursday',
    DaysType.Fr.value: 'Friday',
    DaysType.Sa.value: 'Saturday',
    DaysType.Su.value: 'Sunday'
}

DAYS_PER_LANGUAGE = {
    RegionalSettingsType.ES.value: DAYS_FOR_SP,
    RegionalSettingsType.ES_VE.value: DAYS_FOR_SP,
    RegionalSettingsType.EN.value: DAYS_FOR_EN,
    RegionalSettingsType.EN_US.value: DAYS_FOR_EN,
}

LIST_TRUNCATED_LAST_ITEM_PER_LANGUAGE = {
    RegionalSettingsType.ES.value: 'mas',
    RegionalSettingsType.ES_VE.value: 'mas',
    RegionalSettingsType.EN.value: 'more',
    RegionalSettingsType.EN_US.value: 'more',
}

LIST_LAST_ITEM_SEPARATOR_PER_LANGUAGE = {
    RegionalSettingsType.ES.value: 'y',
    RegionalSettingsType.ES_VE.value: 'y',
    RegionalSettingsType.EN.value: 'and',
    RegionalSettingsType.EN_US.value: 'and',
}



# coding=utf-8
import os
import configparser
from distutils.util import strtobool
from pkg_resources import resource_string

__all__ = [
    "annotations",
    "annotations_validator",
    "basic_tokenizer",
    "corpus",
    "corpus_transformer",
    "google_api",
    "indexer",
    "lexicon",
    "mammutGraph",
    "mammut_api",
    "ngram",
    "package",
    "simulator",
    "ssh_commands",
    "storage",
    "universal_dependencies",
    "util",
]

FEATURES_CONFIGURATIONS = configparser.ConfigParser()
FEATURES_CONFIGURATIONS.read_string(
    resource_string("mammut.resources", "features.ini").decode("utf-8")
)
FEATURES_CONFIGURATIONS.read(["data/features.ini", "features.ini"])

WORD_LEXICAL_RELEVANCE_INDEX_ESTIMATOR_ADD_NEW_SAMPLE = strtobool(
    FEATURES_CONFIGURATIONS["BASIC_TOKENIZER"][
        "word_lexical_relevance_index_estimator_add_new_sample"
    ]
)
GOOGLE_API_AUTH_LOCAL_WEBSERVER = strtobool(
    FEATURES_CONFIGURATIONS["GOOGLE_API"]["auth_local_webserver"]
)
if os.environ.get("GOOGLE_API_AUTH_LOCAL_CREDENTIAL") is not None:
    GOOGLE_API_AUTH_LOCAL_CREDENTIAL = strtobool(
        os.environ.get("GOOGLE_API_AUTH_LOCAL_CREDENTIAL")
    )
else:
    GOOGLE_API_AUTH_LOCAL_CREDENTIAL = strtobool(
        FEATURES_CONFIGURATIONS["GOOGLE_API"]["auth_local_credential"]
    )

global_request_timeout = float(
    FEATURES_CONFIGURATIONS["INDEXING"]["global_request_timeout"]
)
index_creation_timeout = FEATURES_CONFIGURATIONS["INDEXING"]["index_creation_timeout"]
index_definition_timeout = FEATURES_CONFIGURATIONS["INDEXING"][
    "index_definition_timeout"
]
index_lemma_timeout = FEATURES_CONFIGURATIONS["INDEXING"]["index_lemma_timeout"]



from collections import namedtuple
import mammut
from mammut.common.constants import DictionaryTable
from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.storage.storage_manager import StorageManager
from mammut.common.indexer import DefinitionIndexer
import sys
import pandas as pd


class ExternalDictionarySource:
    def __init__(
        self,
        source_id: str,
        source_url_pattern: str,
        html_wrapper: str,
        regional_settings: str,
        standard,
    ):
        self.source_id = source_id
        self.source_url_pattern = source_url_pattern
        self.html_wrapper = html_wrapper
        self.standard = standard
        self.regional_settings = regional_settings

    def get_url(self, lemma: str):
        return self.source_url_pattern.format(lemma)

    def get_iframe(self, lemma: str):
        return self.html_wrapper.format(self.get_url(lemma))


class DictionaryPosTagEntry:
    def __init__(self, pos_tag: str, postag_group_by_object):
        self.pos_tag = pos_tag
        self.postag_group_by_object = postag_group_by_object
        self.definitions_set_entries = []
        self.morpheme_paradigm = ""
        self.misspellings_paradigm = ""
        self.features_paradigm = ""
        if self.postag_group_by_object is not None:
            self.morpheme_paradigm = list(
                self.postag_group_by_object[
                    DictionaryTable.MORPH_PARADIGM_COLUMN_NAME()
                ]
            )[0]
            self.misspellings_paradigm = list(
                self.postag_group_by_object[DictionaryTable.MISS_PARADIGM_COLUMN_NAME()]
            )[0]
            self.features_paradigm = list(
                self.postag_group_by_object[DictionaryTable.FET_PARADIGM_COLUMN_NAME()]
            )[0]
            self.definitions_set_group_by_object = self.postag_group_by_object.groupby(
                [DictionaryTable.DEFINITION_INDEX_COLUMN_NAME()]
            )
            for g in self.definitions_set_group_by_object:
                self.definitions_set_entries.append(
                    DictionaryDefinitionSetEntry(int(g[0]), g[1])
                )

    def get_definitions_count(self):
        return sum(
            [len(dsg.definitions_entries) for dsg in self.definitions_set_entries]
        )


class DictionaryLemmaEntry:
    def __init__(self, lemma: str, lemma_group_by_object):
        self.lemma = lemma
        self.lemma_group_by_object = lemma_group_by_object
        self.postags_entries = []
        self.postags_index_by_postag_id = {}
        self.word_type = ""
        self.regional_settings_lemma = ""
        self.afi = ""
        if self.lemma_group_by_object is not None:
            self.word_type = list(
                self.lemma_group_by_object[DictionaryTable.WORD_TYPE_COLUMN_NAME()]
            )[0]
            self.regional_settings_lemma = list(
                self.lemma_group_by_object[
                    DictionaryTable.REGIONAL_SETTINGS_LEMMA_COLUMN_NAME()
                ]
            )[0]
            self.afi = list(
                self.lemma_group_by_object[DictionaryTable.AFI_COLUMN_NAME()]
            )[0]
            self.postags_group_by_object = self.lemma_group_by_object.groupby(
                [DictionaryTable.POS_PARADIGM_COLUMN_NAME()]
            )
            for g in self.postags_group_by_object:
                self.add_pos_tag_entry(DictionaryPosTagEntry(g[0], g[1]))

    def get_definitions_count(self):
        return sum(
            [
                len(dsg.definitions_entries)
                for pg in self.postags_entries
                for dsg in pg.definitions_set_entries
            ]
        )

    def add_pos_tag_entry(self, pos_tag_entry: DictionaryPosTagEntry):
        self.postags_entries.append(pos_tag_entry)
        index = self.postags_entries.index(pos_tag_entry)
        self.postags_index_by_postag_id[pos_tag_entry.pos_tag] = index

    def remove_pos_tag_entry(self, pos_tag_entry: DictionaryPosTagEntry):
        self.postags_entries.remove(pos_tag_entry)
        del self.postags_index_by_postag_id[pos_tag_entry.pos_tag]


LexicalizationStatistics = namedtuple(
    "LexicalizationStatistics",
    [
        "lemmas_count",
        "definitions_count",
        "morphemes_count",
        "words_count",
        "undefined_words_count",
    ],
)


class DictionaryDefinitionSetEntry:
    def __init__(self, definition_index: int, definition_set_group_by_object):
        self.definition_index = definition_index
        self.definition_set_group_by_object = definition_set_group_by_object
        self.definitions_entries = []
        self.features = ""
        self.features_list = []
        if self.definition_set_group_by_object is not None:
            self.features = list(
                self.definition_set_group_by_object[
                    DictionaryTable.FEATURES_COLUMN_NAME()
                ]
            )[0]
            if self.features:
                self.features_list = self.features.split(",")
            for d in self.definition_set_group_by_object.iterrows():
                self.definitions_entries.append(
                    DictionaryDefinitionEntry(
                        d[1][DictionaryTable.SOURCE_COLUMN_NAME()],
                        d[1][DictionaryTable.TOPIC_COLUMN_NAME()],
                        d[1][DictionaryTable.DEFINITION_COLUMN_NAME()],
                        d[1][
                            DictionaryTable.REGIONAL_SETTINGS_DEFINITION_COLUMN_NAME()
                        ],
                        d[1][DictionaryTable.DATE_COLUMN_NAME()],
                        d[0],
                    )
                )

    def set_features_list(self, features_to_set):
        self.features_list = [fet_desc.feature_id for fet_desc in features_to_set]
        self.features = ""
        length = len(self.features_list)
        for i in range(0, length):
            self.features += self.features_list[i]
            if i < length - 1:
                self.features += ","

    def get_definitions_count(self):
        return len(self.definitions_entries)


class DictionaryDefinitionEntry:
    def __init__(
        self,
        source: str,
        topic: str,
        definition: str,
        regional_settings_definition: str,
        date: str,
        id: int = None,
    ):
        self.source = source
        self.topic = topic
        self.definition = definition
        self.regional_settings_definition = regional_settings_definition
        self.date = date
        self.id = id


class Dictionary:
    from mammut.common.lexicon.linguistic_standard import LinguisticStandard
    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "P"
    DISAMBIGUATION_DEFAULT_END_CELL = "E"
    DISAMBIGUATION_SHEET_TITLE = "disambiguation"

    def __init__(
        self,
        spreadsheet_id: str,
        sheet_title: str,
        synonym_sheet_title: str,
        tokenizer: BasicTokenizer,
        standard: LinguisticStandard,
        is_base: bool = False,
        package_name="",
        local_storage: bool = False,
    ):
        self.package_name = package_name
        if not self.package_name:
            self.package_name = spreadsheet_id.lower()
        self.storage_manager = StorageManager()
        self.definition_indexer = DefinitionIndexer(self.package_name)
        self.standard = standard
        self.tokenizer = tokenizer
        self.tokenizer.set_dictionary(self)
        self.new_dictionaries_count = 0
        self.spreadsheet_id = spreadsheet_id
        self.original_sheet_title = sheet_title
        self.sheet_title = sheet_title
        self.synonym_sheet_title = synonym_sheet_title
        self.disambiguation_sheet_title = Dictionary.DISAMBIGUATION_SHEET_TITLE
        self.range_name = self.sheet_title + "!" + Dictionary.DEFAULT_START_CELL
        self.disambiguation_range_name = (
            self.disambiguation_sheet_title + "!" + Dictionary.DEFAULT_START_CELL
        )
        self.lemmas_entries = []
        self.morphemes = []
        self.lemmas_entries_by_lemma = {}
        self.lemmas_entries_by_morpheme = {}
        self.lemma_layer = {}
        self.synonym_inverted_index = {}
        self.data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.sheet_title,
            Dictionary.DEFAULT_START_COLUMN,
            Dictionary.DEFAULT_START_ROW,
            Dictionary.DEFAULT_END_CELL,
        )

        self.definition_indexer.index_dictionary(
            self.data_frame, self.standard, self.spreadsheet_id, self.sheet_title
        )
        if not is_base:
            self.disambiguation_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
                self.spreadsheet_id,
                self.disambiguation_sheet_title,
                Dictionary.DEFAULT_START_COLUMN,
                Dictionary.DEFAULT_START_ROW,
                Dictionary.DISAMBIGUATION_DEFAULT_END_CELL,
            )

            self.standard.set_working_dictionary(self)
        if self.data_frame is not None:
            self.id_count = len(self.data_frame)
            self.lemmas_group_by_object = self.data_frame.groupby(
                [DictionaryTable.LEMMA_COLUMN_NAME()]
            )
            for g in self.lemmas_group_by_object:
                lemma_entry = DictionaryLemmaEntry(g[0], g[1])
                self.add_lemma_entry(lemma_entry)
        else:
            self.id_count = 0
        self.synonym_range_name = (
            self.synonym_sheet_title + "!" + Dictionary.DEFAULT_START_CELL
        )
        self.synonym_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.synonym_sheet_title,
            Dictionary.DEFAULT_START_COLUMN,
            Dictionary.DEFAULT_START_ROW,
            Dictionary.DEFAULT_END_CELL,
        )

        if self.synonym_data_frame is not None:
            for r in self.synonym_data_frame.iterrows():
                sense_lemma = r[1][DictionaryTable.SENSE_LEMMA_COLUMN_NAME()]
                syn_words = r[1][DictionaryTable.SYNONYM_WORDS_COLUMN_NAME()]
                for s_w in syn_words.split(","):
                    s_w = s_w.strip().lower()
                    if s_w in self.synonym_inverted_index:
                        self.synonym_inverted_index[s_w].append(sense_lemma)
                    else:
                        self.synonym_inverted_index[s_w] = [sense_lemma]

    def contains_lemma(self, lemma: str):
        return lemma in self.lemmas_entries_by_lemma

    def contains_morpheme(self, morpheme: str):
        return morpheme in self.lemmas_entries_by_morpheme

    def contains_synonym(self, word: str):
        return word in self.synonym_inverted_index

    def _update_lemma_entries(self, lemmas, regional_settings):
        entries = []
        receipts = ""
        for lg in lemmas:
            for pg in lg.postags_entries:
                morphemes = self.standard.get_morphemes(
                    pg.pos_tag,
                    pg.morpheme_paradigm.split(" ")[0],
                    lg.lemma,
                    regional_settings,
                )
                for dsg in pg.definitions_set_entries:
                    for d in dsg.definitions_entries:
                        entry = [
                            lg.lemma,
                            lg.word_type,
                            lg.regional_settings_lemma,
                            lg.afi,
                            pg.pos_tag,
                            pg.morpheme_paradigm,
                            pg.misspellings_paradigm,
                            pg.features_paradigm,
                            dsg.definition_index,
                            dsg.features,
                            d.source,
                            d.topic,
                            d.definition,
                            d.regional_settings_definition,
                            d.date,
                        ]
                        entries.append(entry)
                        definition_to_index = entry
                        self.definition_indexer.index_lemma(
                            definition_to_index, morphemes
                        )
                        if d.id is not None:
                            self.storage_manager.update_row_from_spreadsheet(
                                self.spreadsheet_id,
                                self.sheet_title,
                                entry,
                                "A",
                                str(d.id + 2),
                            )
                        else:
                            self.storage_manager.append_row_to_sheet(
                                self.spreadsheet_id, "dictionary", entry
                            )
                            next_id = self.id_count
                            d.id = next_id
                            self.id_count = self.id_count + 1

        self.data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.sheet_title,
            Dictionary.DEFAULT_START_COLUMN,
            Dictionary.DEFAULT_START_ROW,
            Dictionary.DEFAULT_END_CELL,
        )

        for e in entries:
            if receipts:
                receipts += "\n"
            receipts += self.definition_indexer.get_receipt(e[0], e[12])
        return receipts

    def delete_definition_lemma(self, lemma_to_delete, definition):
        data = []
        data.append(
            [
                DictionaryTable.LEMMA_COLUMN_NAME(),
                DictionaryTable.WORD_TYPE_COLUMN_NAME(),
                DictionaryTable.REGIONAL_SETTINGS_LEMMA_COLUMN_NAME(),
                DictionaryTable.AFI_COLUMN_NAME(),
                DictionaryTable.POS_PARADIGM_COLUMN_NAME(),
                DictionaryTable.MORPH_PARADIGM_COLUMN_NAME(),
                DictionaryTable.MISS_PARADIGM_COLUMN_NAME(),
                DictionaryTable.FET_PARADIGM_COLUMN_NAME(),
                DictionaryTable.DEFINITION_INDEX_COLUMN_NAME(),
                DictionaryTable.FEATURES_COLUMN_NAME(),
                DictionaryTable.SOURCE_COLUMN_NAME(),
                DictionaryTable.TOPIC_COLUMN_NAME(),
                DictionaryTable.DEFINITION_COLUMN_NAME(),
                DictionaryTable.REGIONAL_SETTINGS_DEFINITION_COLUMN_NAME(),
                DictionaryTable.DATE_COLUMN_NAME(),
                DictionaryTable.TIMESTAMP_COLUMN_NAME(),
            ]
        )
        deleted = False
        for index, lemma in enumerate(
            self.data_frame[DictionaryTable.LEMMA_COLUMN_NAME()]
        ):
            if (
                lemma == lemma_to_delete
                and self.data_frame[DictionaryTable.DEFINITION_COLUMN_NAME()][index]
                == definition
            ):
                deleted = True
                pass
            else:
                if lemma == lemma_to_delete and deleted:
                    self.data_frame[DictionaryTable.DEFINITION_INDEX_COLUMN_NAME()][
                        index
                    ] = str(
                        int(
                            self.data_frame[
                                DictionaryTable.DEFINITION_INDEX_COLUMN_NAME()
                            ][index]
                        )
                        - 1
                    )
                data.append(
                    [
                        self.data_frame[DictionaryTable.LEMMA_COLUMN_NAME()][index],
                        self.data_frame[DictionaryTable.WORD_TYPE_COLUMN_NAME()][index],
                        self.data_frame[
                            DictionaryTable.REGIONAL_SETTINGS_LEMMA_COLUMN_NAME()
                        ][index],
                        self.data_frame[DictionaryTable.AFI_COLUMN_NAME()][index],
                        self.data_frame[DictionaryTable.POS_PARADIGM_COLUMN_NAME()][
                            index
                        ],
                        self.data_frame[DictionaryTable.MORPH_PARADIGM_COLUMN_NAME()][
                            index
                        ],
                        self.data_frame[DictionaryTable.MISS_PARADIGM_COLUMN_NAME()][
                            index
                        ],
                        self.data_frame[DictionaryTable.FET_PARADIGM_COLUMN_NAME()][
                            index
                        ],
                        int(
                            self.data_frame[
                                DictionaryTable.DEFINITION_INDEX_COLUMN_NAME()
                            ][index]
                        ),
                        self.data_frame[DictionaryTable.FEATURES_COLUMN_NAME()][index],
                        self.data_frame[DictionaryTable.SOURCE_COLUMN_NAME()][index],
                        self.data_frame[DictionaryTable.TOPIC_COLUMN_NAME()][index],
                        self.data_frame[DictionaryTable.DEFINITION_COLUMN_NAME()][
                            index
                        ],
                        self.data_frame[
                            DictionaryTable.REGIONAL_SETTINGS_DEFINITION_COLUMN_NAME()
                        ][index],
                        self.data_frame[DictionaryTable.DATE_COLUMN_NAME()][index],
                        self.data_frame[DictionaryTable.TIMESTAMP_COLUMN_NAME()][index],
                    ]
                )

        self.storage_manager.delete_sheet(
            self.spreadsheet_id,
            self.storage_manager.get_sheet_id(self.spreadsheet_id, "dictionary"),
        )
        self.storage_manager.create_new_sheet(self.spreadsheet_id, "dictionary")
        self.storage_manager.add_rows_to_spreadsheet(
            self.spreadsheet_id, self.sheet_title + "!" + "A1", data
        )
        self.data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.sheet_title,
            Dictionary.DEFAULT_START_COLUMN,
            Dictionary.DEFAULT_START_ROW,
            Dictionary.DEFAULT_END_CELL,
        )

    def _save_lemma_entries(self, lemmas, regional_settings):
        entries = []
        receipts = ""
        for lg in lemmas:
            for pg in lg.postags_entries:
                morphemes = self.standard.get_morphemes(
                    pg.pos_tag,
                    pg.morpheme_paradigm.split(" ")[0],
                    lg.lemma,
                    regional_settings,
                )
                for dsg in pg.definitions_set_entries:
                    for d in dsg.definitions_entries:
                        next_id = (
                            self.id_count
                        )  # TODO:esto de setear los id's deberia ir el los metodos privados??
                        d.id = next_id
                        self.id_count = self.id_count + 1
                        entry = [
                            lg.lemma,
                            lg.word_type,
                            lg.regional_settings_lemma,
                            lg.afi,
                            pg.pos_tag,
                            pg.morpheme_paradigm,
                            pg.misspellings_paradigm,
                            pg.features_paradigm,
                            dsg.definition_index,
                            dsg.features,
                            d.source,
                            d.topic,
                            d.definition,
                            d.regional_settings_definition,
                            d.date,
                            str(mammut.get_timezone_aware_datetime_now()),
                        ]
                        definition_to_index = entry
                        self.definition_indexer.index_lemma(
                            definition_to_index, morphemes
                        )
                        entries.append(entry)
        self.storage_manager.add_rows_to_spreadsheet(
            self.spreadsheet_id, self.range_name, entries
        )
        self.data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.sheet_title,
            Dictionary.DEFAULT_START_COLUMN,
            Dictionary.DEFAULT_START_ROW,
            Dictionary.DEFAULT_END_CELL,
        )

        for e in entries:
            if receipts:
                receipts += "\n"
            receipts += self.definition_indexer.get_receipt(e[0], e[12])
        return receipts

    def add_lemma_entry(
        self, lemma_entry: DictionaryLemmaEntry, save_to_google_sheet=False
    ):
        regional_settings = lemma_entry.regional_settings_lemma
        result = None
        self.lemmas_entries_by_lemma[lemma_entry.lemma] = lemma_entry
        self.lemmas_entries.append(lemma_entry)
        for pg in lemma_entry.postags_entries:
            layer = 0
            for m in self.standard.get_morphemes(
                pg.pos_tag, pg.morpheme_paradigm, lemma_entry.lemma, regional_settings
            ):
                if m.text not in self.lemmas_entries_by_morpheme:
                    self.lemmas_entries_by_morpheme[m.text] = []
                    self.morphemes.append(m.text)
                self.lemmas_entries_by_morpheme[m.text].append(lemma_entry)
                if (
                    m.text in self.tokenizer.stats[regional_settings].word_layer
                    and self.tokenizer.stats[regional_settings].word_layer[m.text]
                    > layer
                ):
                    layer = self.tokenizer.stats[regional_settings].word_layer[m.text]
            for dsg in pg.definitions_set_entries:
                for d in dsg.definitions_entries:
                    self.tokenizer.tokenize_definition(
                        d.definition, layer + 1, regional_settings, None
                    )
        if save_to_google_sheet:
            result = self._save_lemma_entries([lemma_entry], regional_settings)
        return result

    def update_lemma_entry(
        self, lemma_entry: DictionaryLemmaEntry, save_to_google_sheet=False
    ):
        regional_settings = lemma_entry.regional_settings_lemma
        result = None
        self.lemmas_entries_by_lemma[lemma_entry.lemma] = lemma_entry
        old_lemma_entry_index = next(
            index
            for (index, d) in enumerate(self.lemmas_entries)
            if d.lemma == lemma_entry.lemma
        )
        self.lemmas_entries[old_lemma_entry_index] = lemma_entry

        for pg in lemma_entry.postags_entries:
            layer = 0
            for m in self.standard.get_morphemes(
                pg.pos_tag, pg.morpheme_paradigm, lemma_entry.lemma, regional_settings
            ):
                if m.text not in self.lemmas_entries_by_morpheme:
                    self.lemmas_entries_by_morpheme[m.text] = []
                    self.morphemes.append(m.text)
                self.lemmas_entries_by_morpheme[m.text].append(lemma_entry)
                if (
                    m.text in self.tokenizer.stats[regional_settings].word_layer
                    and self.tokenizer.stats[regional_settings].word_layer[m.text]
                    > layer
                ):
                    layer = self.tokenizer.stats[regional_settings].word_layer[m.text]
            for dsg in pg.definitions_set_entries:
                for d in dsg.definitions_entries:
                    self.tokenizer.tokenize_definition(
                        d.definition, layer + 1, regional_settings, None
                    )
        if save_to_google_sheet:
            result = self._update_lemma_entries([lemma_entry], regional_settings)
        return result

    def create_new_dictionary(self, regional_settings):
        self.new_dictionaries_count += 1
        self.sheet_title = self.original_sheet_title + str(self.new_dictionaries_count)
        self.range_name = self.sheet_title + "!" + Dictionary.DEFAULT_START_CELL
        self.storage_manager.create_new_sheet(self.spreadsheet_id, self.sheet_title)
        result = self._save_lemma_entries(self.lemmas_entries, regional_settings)
        return result

    def get_undefined_morphemes_generator(
        self, regional_settings, layer=0, with_index=False
    ):
        return self.get_morphemes_generator(regional_settings, layer, with_index, True)

    def get_defined_morphemes_generator(
        self, regional_settings, layer=0, with_index=False
    ):
        return self.get_morphemes_generator(regional_settings, layer, with_index, False)

    def get_morphemes_generator(self, regional_settings, layer, with_index, undefined):
        stats_words_generator = self.tokenizer.stats[
            regional_settings
        ].get_words_generator(layer, with_index)
        stop = False
        while not stop:
            try:
                g_r = next(stats_words_generator)
                if with_index:
                    w = g_r[0]
                else:
                    w = g_r
                if undefined:
                    if (
                        not self.contains_morpheme(w)
                        and not self.standard.base_dictionary.contains_morpheme(w)
                        and not self.standard.has_functionals_by_lemma_regional_settings(
                            w, regional_settings
                        )
                        and not self.contains_synonym(w)
                        and not self.standard.base_dictionary.contains_synonym(w)
                    ):
                        yield g_r
                elif not undefined:
                    if (
                        self.contains_morpheme(w)
                        or self.standard.base_dictionary.contains_morpheme(w)
                        or self.standard.has_functionals_by_lemma_regional_settings(
                            w, regional_settings
                        )
                        or self.contains_synonym(w)
                        or self.standard.base_dictionary.contains_synonym(w)
                    ):
                        yield g_r
            except StopIteration:
                stop = True

    def get_definitions_count(self):
        return sum(
            [
                len(dsg.definitions_entries)
                for lg in self.lemmas_entries
                for pg in lg.postags_entries
                for dsg in pg.definitions_set_entries
            ]
        )

    def get_morphemes_count(self):
        return len(self.lemmas_entries_by_morpheme)

    def get_undefined_morphemes_count(self, regional_settings):
        return sum(
            [
                1
                if not self.contains_morpheme(w)
                and not self.standard.has_functionals_by_lemma_regional_settings(
                    w, regional_settings
                )
                else 0
                for w in self.tokenizer.stats[regional_settings].get_words_generator(
                    sys.maxsize
                )
            ]
        )

    def get_lemma(self, lemma: str):
        if self.contains_lemma(lemma):
            lemma_result = self.lemmas_entries_by_lemma[lemma]
        else:
            lemma_result = None
        return lemma_result

    def get_morphemes_of_lemma(self, lemma: str, as_tuple: bool = False
                               ):
        result = set()
        lemma_entry = self.get_lemma(lemma)
        if lemma_entry is not None:
            regional_settings = lemma_entry.regional_settings_lemma
            for pg in lemma_entry.postags_entries:
                for m in self.standard.get_morphemes(
                    pg.pos_tag,
                    pg.morpheme_paradigm,
                    lemma_entry.lemma,
                    regional_settings,
                ):
                    result.add(m.text)
        return result

    def get_implicit_functional_lemma_name(
        self, lemma, textbound_annotation, reference_dict
    ):
        functional_visual_identifier = "*"
        print("functional implicit lemma name", lemma)
        name = "{functional}({lemma})*".format(
            functional=lemma, lemma=textbound_annotation.text
        )
        if name in reference_dict.keys():
            print("se esta repitiendo un value en el dict")
        return name

    def get_functional_lemma_name(self, lemma, index, reference_dict):
        print("functional lemma name", lemma)
        name = "{functional}({index})*".format(functional=lemma, index=index)
        if name in reference_dict.keys():
            print("se esta repitiendo un value en el dict")
        return name

    def get_index_from_span(self, event_annotation, textbound_annotation):
        document = event_annotation.documents_by_dimension["t-pos"]
        sorted_spans = [
            tbsorted.spans[0]
            for tbsorted in sorted(
                document.ann_obj.get_textbounds(), key=lambda tb: tb.spans
            )
        ]
        return sorted_spans.index(textbound_annotation.spans[0])

    def get_functional_disambiguation_dicts(
        self,
        event_annotation,
        textbound_annotation,
        tokens_dict,
        tokens_info_dict,
        tokens_type_resume,
        implicit_functional_lemma_descriptor=None,
    ):
        textbound_index = self.get_index_from_span(
            event_annotation, textbound_annotation
        )
        # TODO: aqui debe ser el lemma de la palabra funcional cuando es implicita
        if implicit_functional_lemma_descriptor is not None:
            tokens_type_resume["functional"].append(
                implicit_functional_lemma_descriptor.lemma
            )
            funtional_lemma_descriptor = implicit_functional_lemma_descriptor
            span = ""
            dict_key = self.get_implicit_functional_lemma_name(
                implicit_functional_lemma_descriptor.lemma,
                textbound_annotation,
                tokens_dict,
            )
            tokens_dict[
                dict_key
            ] = implicit_functional_lemma_descriptor.implicit_condition_id
            tokens_info_dict[dict_key] = [
                event_annotation.get_id(),
                textbound_annotation.text,
                span,
                textbound_annotation.type,
            ]
        else:
            tokens_type_resume["functional"].append(textbound_annotation.text.lower())
            funtional_lemma_descriptor = self.standard.get_functional_lemma_descriptor(
                textbound_annotation.text.lower(),
                event_annotation.event["regnional_settings"],
                textbound_annotation.type,
            )
            span = (
                str(textbound_annotation.spans[0][0])
                + ","
                + str(textbound_annotation.spans[0][1])
            )
            dict_key = self.get_functional_lemma_name(
                textbound_annotation.text.lower(), textbound_index, tokens_dict
            )
            tokens_dict[dict_key] = (
                funtional_lemma_descriptor.lemma
                + "-"
                + event_annotation.event["regnional_settings"]
                + "-"
                + textbound_annotation.type
            )
            tokens_info_dict[dict_key] = [
                event_annotation.get_id(),
                textbound_annotation.text,
                span,
                textbound_annotation.type,
            ]
        return tokens_dict, tokens_info_dict, tokens_type_resume

    def get_event_tokens_dict_and_info(
        self, event_annotation
    ):  # TODO definir bien el nombre
        tokens_dict = {}
        tokens_info_dict = {}
        tokens_type_resume = {"lexical": [], "functional": [], "none": []}
        document = event_annotation.documents_by_dimension["t-pos"]
        ZERO_ARTICLE = "zero_article"
        for i in document.ann_obj.get_textbounds():
            attr_anns = [
                attribute_annotation
                for attribute_annotation in document.ann_obj.get_attributes()
                if i.id == attribute_annotation.target
                and attribute_annotation.value == ZERO_ARTICLE
            ]
            if i.text.lower() in self.lemmas_entries_by_morpheme:
                tokens_type_resume["lexical"].append(i.text.lower())
                for pos_index, pos_tag_entry in enumerate(
                    self.lemmas_entries_by_morpheme[i.text.lower()][0].postags_entries
                ):
                    # TODO: aqui te falta un gran if que condicione todo la ejecucion de toda la logica dentro del for
                    # la logica dentro del for solo te interesa si el pos-tag de la iteracion coincide con el pos-tag en la anotacion
                    if (
                        i.type == pos_tag_entry.pos_tag
                        or i.type
                        in self.standard.pos_descriptor_list[
                            pos_tag_entry.pos_tag
                        ].pos_shifts_ids
                    ):
                        if (
                            len(
                                self.lemmas_entries_by_morpheme[i.text.lower()][0]
                                .postags_entries[pos_index]
                                .definitions_set_entries
                            )
                            > 1
                        ):
                            tokens_dict[i.text.lower()] = (
                                self.lemmas_entries_by_morpheme[i.text.lower()][0].lemma
                                + "-"
                                + pos_tag_entry.pos_tag
                            )
                            # TODO deberia crear metodo para generar esta columan, lo hago aqui para no volver a iterar sobre los textbounbds, revisar que sale mejor
                            tokens_info_dict[i.text.lower()] = [
                                event_annotation.get_id(),
                                i.text,
                                str(i.spans[0][0]) + "," + str(i.spans[0][1]),
                                i.type,
                            ]
                        # TODO: debemos tener un if que siempre revise todas las anotaciones de todas las palabras lexicas
                        # para validar si tiene algun feature que indica la presencia de una palabra funcional implicita
                        # aqui abajo debes pasar el feature class y el feature id, no se como sacarlos de la anotacion
                        # el pos-tag y el regional settings creo que si lo pase bien
                        if len(attr_anns):
                            feature_pos, feature_class = attr_anns[0].type.split("-")
                            feature_id = attr_anns[0].value
                            implicit_functional_lemma_descriptor = self.standard.get_implicit_functional_lemma_descriptor(
                                feature_class,
                                feature_id,
                                feature_pos,
                                event_annotation.event["regnional_settings"],
                            )
                            if implicit_functional_lemma_descriptor is not None:
                                print("###################")
                                print(
                                    "el implicit functional es:",
                                    implicit_functional_lemma_descriptor.lemma,
                                )
                                # TODO: debes hacer algo parecido a lo que haces abajo cuando es un funcional
                                (
                                    tokens_dict,
                                    tokens_info_dict,
                                    tokens_type_resume,
                                ) = self.get_functional_disambiguation_dicts(
                                    event_annotation,
                                    i,
                                    tokens_dict,
                                    tokens_info_dict,
                                    tokens_type_resume,
                                    implicit_functional_lemma_descriptor,
                                )
            elif (
                self.standard.get_functional_lemma_descriptor(
                    i.text.lower(), event_annotation.event["regnional_settings"], i.type
                )
                is not None
            ):
                (
                    tokens_dict,
                    tokens_info_dict,
                    tokens_type_resume,
                ) = self.get_functional_disambiguation_dicts(
                    event_annotation,
                    i,
                    tokens_dict,
                    tokens_info_dict,
                    tokens_type_resume,
                )
            else:
                tokens_type_resume["none"].append(i.text.lower())
        return tokens_dict, tokens_info_dict, tokens_type_resume

    def get_definitions_entries_as_list(self, definitions_set_entries_id):
        id_elements = definitions_set_entries_id.split("-")
        lemma = id_elements[0]
        if lemma in self.lemmas_entries_by_lemma:
            pos = id_elements[1]
            pos_index = self.lemmas_entries_by_lemma[lemma].postags_index_by_postag_id[
                pos
            ]
            definition_set = []
            for i in (
                self.lemmas_entries_by_lemma[lemma]
                .postags_entries[pos_index]
                .definitions_set_entries
            ):
                definitions = []
                for j in i.definitions_entries:
                    definitions.append(j.definition)
                definition_set.append(definitions)
            return definition_set
        else:
            definition_set = []
            functional_lemma_descriptor = None
            if len(id_elements) == 3:  # id normal de lemma descriptor
                regional_settings = id_elements[1]
                pos = id_elements[2]
                # TODO: como hacemos para pedir los implicit_functional_lamadescripotors?pasamos un param que digasi buscamos functional o funtional implicit?
                if self.standard.get_functional_lemma_descriptor(
                    lemma, regional_settings, pos
                ):
                    functional_lemma_descriptor = self.standard.get_functional_lemma_descriptor(
                        lemma, regional_settings, pos
                    )
            elif len(id_elements) == 4:  # implicit_condition_id
                feature_class = id_elements[0]
                feature_id = id_elements[1]
                feature_pos = id_elements[2]
                regional_settings_lemma = id_elements[3]
                if self.standard.get_implicit_functional_lemma_descriptor(
                    feature_class, feature_id, feature_pos, regional_settings_lemma
                ):
                    functional_lemma_descriptor = self.standard.get_implicit_functional_lemma_descriptor(
                        feature_class, feature_id, feature_pos, regional_settings_lemma
                    )
            else:
                raise RuntimeError("Invalid functional lemma id.")
            for i in functional_lemma_descriptor.definitions:
                definition_set.append([i.get_description()])
            return definition_set

    def get_definition_complete_id(self, definitions_set_entries_id, index):
        splited_id = definitions_set_entries_id.split("-")
        if len(splited_id) == 2:
            return definitions_set_entries_id + "-" + str(index)
        elif len(splited_id) == 3:
            descriptor = self.standard.get_functional_lemma_descriptor(
                splited_id[0], splited_id[1], splited_id[2]
            )
            definition = descriptor.definitions[index]
            return (
                definition.lemma
                + "-"
                + descriptor.pos_paradigm
                + "-"
                + definition.index
            )
        elif len(splited_id) == 4:
            descriptor = self.standard.get_implicit_functional_lemma_descriptor(
                splited_id[0], splited_id[1], splited_id[2], splited_id[3]
            )
            definition = descriptor.definitions[index]
            return (
                definition.lemma
                + "-"
                + descriptor.pos_paradigm
                + "-"
                + definition.index
            )

    def is_disambiguation_selected(self, token_info, result_id):
        row_in_data_frame = self.disambiguation_data_frame.query(
            'id == "'
            + token_info[0]
            + '" and text == "'
            + token_info[1]
            + '" and span == "'
            + token_info[2]
            + '" and pos == "'
            + token_info[3]
            + '"  and definition_id == "'
            + result_id
            + '"'
        )
        return len(row_in_data_frame) > 0

    def save_lema_disambiguation(self, token_info, definition_id):
        row_in_data_frame = self.disambiguation_data_frame.query(
            'id == "'
            + token_info[0]
            + '" and text == "'
            + token_info[1]
            + '" and span == "'
            + token_info[2]
            + '" and pos == "'
            + token_info[3]
            + '"'
        )
        if len(row_in_data_frame) > 0:
            index = row_in_data_frame.index.values[0]
            self.storage_manager.delete_rows_from_spreadsheet(
                self.spreadsheet_id,
                Dictionary.DISAMBIGUATION_SHEET_TITLE,
                str(index + 1),
                str(index + 2),
            )
            self.disambiguation_data_frame = self.disambiguation_data_frame.drop(index)
        row = token_info + [definition_id]
        self.disambiguation_data_frame = self.disambiguation_data_frame.append(
            pd.Series(row, list(self.disambiguation_data_frame.columns)),
            ignore_index=True,
        )
        return self.storage_manager.append_row_to_sheet(
            self.spreadsheet_id, Dictionary.DISAMBIGUATION_SHEET_TITLE, row
        )

    def get_lexicalization_statistics(self, regional_settings: str):
        lemmas_count = 0
        definitions_count = 0
        for lg in self.lemmas_entries:
            lemmas_count = lemmas_count + 1
            for pg in lg.postags_entries:
                for dsg in pg.definitions_set_entries:
                    definitions_count = definitions_count + len(dsg.definitions_entries)
        morphemes_count = self.get_morphemes_count()
        words_count = 0
        undefined_morphemes_count = 0
        stats_words_generator = self.tokenizer.stats[
            regional_settings
        ].get_words_generator(sys.maxsize)
        stop = False
        while not stop:
            try:
                w = next(stats_words_generator)
                words_count = words_count + 1
                if not self.contains_morpheme(
                    w
                ) and not self.standard.has_functionals_by_lemma_regional_settings(
                    w, regional_settings
                ):
                    undefined_morphemes_count = undefined_morphemes_count + 1
            except StopIteration:
                stop = True
        return LexicalizationStatistics(
            lemmas_count,
            definitions_count,
            morphemes_count,
            words_count,
            undefined_morphemes_count,
        )

    def get_words_in_definitions_out_of_vocabulary(self, regional_settings: str):
        res = set()
        for lg in self.lemmas_entries:
            for pg in lg.postags_entries:
                for dsg in pg.definitions_set_entries:
                    for d in dsg.definitions_entries:
                        dov = set(
                            self.standard.is_in_vocabulary(
                                d.definition, regional_settings
                            ).list
                        )
                        res = res.union(dov)
        return res



from mammut.common.indexer import NGramIndexer
import smtplib
from mammut.data.download import *
import os
import time


class Agenda:
    def __init__(self):
        self.ngram_indexer = NGramIndexer()
        self.ngram_data_path = ""
        self.set1_chars = [
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "a",
            "b",
            "c",
            "d",
            "e",
            "f",
            "g",
            "h",
            "i",
            "j",
            "k",
            "l",
            "m",
            "n",
            "o",
            "other",
            "p",
            "pos",
            "punctuation",
            "q",
            "r",
            "s",
            "t",
            "u",
            "v",
            "w",
            "x",
            "y",
            "z",
        ]

        self.set2_chars_1 = [
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "_ADJ_",
            "_ADP_",
            "_ADV_",
            "_CONJ_",
            "_DET_",
            "_NOUN_",
            "_NUM_",
            "_PRON_",
            "_PRT_",
            "_VERB_",
            "a_",
            "aa",
            "ab",
            "ac",
            "ad",
            "ae",
            "af",
            "ag",
            "ah",
            "ai",
            "aj",
            "ak",
            "al",
            "am",
            "an",
            "ao",
            "ap",
            "aq",
            "ar",
            "as",
            "at",
            "au",
            "av",
            "aw",
            "ax",
            "ay",
            "az",
            "b_",
        ]
        self.set2_chars_2 = [
            "ba",
            "bb",
            "bc",
            "bd",
            "be",
            "bf",
            "bg",
            "bh",
            "bi",
            "bj",
            "bk",
            "bl",
            "bm",
            "bn",
            "bo",
            "bp",
            "bq",
            "br",
            "bs",
            "bt",
            "bu",
            "bv",
            "bw",
            "bx",
            "by",
            "bz",
            "c_",
            "ca",
            "cb",
            "cc",
            "cd",
            "ce",
            "cf",
            "cg",
            "ch",
            "ci",
            "cj",
            "ck",
            "cl",
            "cm",
            "cn",
            "co",
            "cp",
            "cq",
            "cr",
            "cs",
            "ct",
            "cu",
            "cv",
            "cw",
            "cx",
            "cy",
            "cz",
            "d_",
            "da",
            "db",
        ]
        self.set2_chars_3 = [
            "dc",
            "dd",
            "de",
            "df",
            "dg",
            "dh",
            "di",
            "dj",
            "dk",
            "dl",
            "dm",
            "dn",
            "do",
            "dp",
            "dq",
            "dr",
            "ds",
            "dt",
            "du",
            "dv",
            "dw",
            "dx",
            "dy",
            "dz",
            "e_",
            "ea",
            "eb",
            "ec",
            "ed",
            "ee",
            "ef",
            "eg",
            "eh",
            "ei",
            "ej",
            "ek",
            "el",
            "em",
            "en",
            "eo",
            "ep",
            "eq",
            "er",
            "es",
            "et",
            "eu",
            "ev",
            "ew",
            "ex",
            "ey",
            "ez",
            "f_",
        ]
        self.set2_chars_4 = [
            "fa",
            "fb",
            "fc",
            "fd",
            "fe",
            "ff",
            "fg",
            "fh",
            "fi",
            "fj",
            "fk",
            "fl",
            "fm",
            "fn",
            "fo",
            "fp",
            "fq",
            "fr",
            "fs",
            "ft",
            "fu",
            "fv",
            "fw",
            "fx",
            "fy",
            "fz",
            "g_",
            "ga",
            "gb",
            "gc",
            "gd",
            "ge",
            "gf",
            "gg",
            "gh",
            "gi",
            "gj",
            "gk",
            "gl",
            "gm",
            "gn",
            "go",
            "gp",
            "gq",
            "gr",
            "gs",
            "gt",
            "gu",
            "gv",
            "gw",
            "gx",
            "gy",
            "gz",
            "h_",
            "ha",
            "hb",
        ]
        self.set2_chars_5 = [
            "hc",
            "hd",
            "he",
            "hf",
            "hg",
            "hh",
            "hi",
            "hj",
            "hk",
            "hl",
            "hm",
            "hn",
            "ho",
            "hp",
            "hq",
            "hr",
            "hs",
            "ht",
            "hu",
            "hv",
            "hw",
            "hx",
            "hy",
            "hz",
            "i_",
            "ia",
            "ib",
            "ic",
            "id",
            "ie",
            "if",
            "ig",
            "ih",
            "ii",
            "ij",
            "ik",
            "il",
            "im",
            "in",
            "io",
            "ip",
            "iq",
            "ir",
            "is",
            "it",
            "iu",
            "iv",
            "iw",
            "ix",
            "iy",
            "iz",
            "j_",
            "ja",
            "jb",
        ]
        self.set2_chars_6 = [
            "jc",
            "jd",
            "je",
            "jf",
            "jg",
            "jh",
            "ji",
            "jj",
            "jk",
            "jl",
            "jm",
            "jn",
            "jo",
            "jp",
            "jq",
            "jr",
            "js",
            "jt",
            "ju",
            "jv",
            "jw",
            "jx",
            "jy",
            "jz",
            "k_",
            "ka",
            "kb",
            "kc",
            "kd",
            "ke",
            "kf",
            "kg",
            "kh",
            "ki",
            "kj",
            "kk",
            "kl",
            "km",
            "kn",
            "ko",
            "kp",
            "kq",
            "kr",
            "ks",
            "kt",
            "ku",
            "kv",
            "kw",
            "kx",
            "ky",
            "kz",
            "l_",
            "la",
            "lb",
        ]
        self.set2_chars_7 = [
            "lc",
            "ld",
            "le",
            "lf",
            "lg",
            "lh",
            "li",
            "lj",
            "lk",
            "ll",
            "lm",
            "ln",
            "lo",
            "lp",
            "lq",
            "lr",
            "ls",
            "lt",
            "lu",
            "lv",
            "lw",
            "lx",
            "ly",
            "lz",
            "m_",
            "ma",
            "mb",
            "mc",
            "md",
            "me",
            "mf",
            "mg",
            "mh",
            "mi",
            "mj",
            "mk",
            "ml",
            "mm",
            "mn",
            "mo",
            "mp",
            "mq",
            "mr",
            "ms",
            "mt",
            "mu",
            "mv",
            "mw",
            "mx",
            "my",
            "mz",
            "n_",
            "na",
            "nb",
        ]
        self.set2_chars_8 = [
            "nc",
            "nd",
            "ne",
            "nf",
            "ng",
            "nh",
            "ni",
            "nj",
            "nk",
            "nl",
            "nm",
            "nn",
            "no",
            "np",
            "nq",
            "nr",
            "ns",
            "nt",
            "nu",
            "nv",
            "nw",
            "nx",
            "ny",
            "nz",
            "o_",
            "oa",
            "ob",
            "oc",
            "od",
            "oe",
            "of",
            "og",
            "oh",
            "oi",
            "oj",
            "ok",
            "ol",
            "om",
            "on",
            "oo",
            "op",
            "oq",
            "or",
            "os",
            "ot",
            "other",
            "ou",
            "ov",
            "ow",
            "ox",
            "oy",
            "oz",
            "p_",
        ]
        self.set2_chars_9 = [
            "pa",
            "pb",
            "pc",
            "pd",
            "pe",
            "pf",
            "pg",
            "ph",
            "pi",
            "pj",
            "pk",
            "pl",
            "pm",
            "pn",
            "po",
            "pp",
            "pq",
            "pr",
            "ps",
            "pt",
            "pu",
            "punctuation",
            "pv",
            "pw",
            "px",
            "py",
            "pz",
            "q_",
            "qa",
            "qb",
            "qc",
            "qd",
            "qe",
            "qf",
            "qg",
            "qh",
            "qi",
            "qj",
        ]
        self.set2_chars_10 = ["qk"]
        self.set2_chars_11 = [
            "ql",
            "qm",
            "qn",
            "qo",
            "qp",
            "qq",
            "qr",
            "qs",
            "qt",
            "qu",
            "qv",
            "qw",
            "qx",
            "qy",
        ]
        self.set2_chars_12 = [
            "qz",
            "r_",
            "ra",
            "rb",
            "rc",
            "rd",
            "re",
            "rf",
            "rg",
            "rh",
            "ri",
            "rj",
            "rk",
            "rl",
            "rm",
            "rn",
            "ro",
            "rp",
            "rq",
            "rr",
            "rs",
            "rt",
            "ru",
            "rv",
            "rw",
            "rx",
            "ry",
            "rz",
            "s_",
            "sa",
            "sb",
            "sc",
            "sd",
            "se",
            "sf",
            "sg",
            "sh",
            "si",
            "sj",
            "sk",
            "sl",
            "sm",
            "sn",
            "so",
            "sp",
            "sq",
            "sr",
            "ss",
            "st",
            "su",
            "sv",
            "sw",
            "sx",
            "sy",
        ]
        self.set2_chars_13 = [
            "sz",
            "t_",
            "ta",
            "tb",
            "tc",
            "td",
            "te",
            "tf",
            "tg",
            "th",
            "ti",
            "tj",
            "tk",
            "tl",
            "tm",
            "tn",
            "to",
            "tp",
            "tq",
            "tr",
            "ts",
            "tt",
            "tu",
            "tv",
            "tw",
            "tx",
            "ty",
            "tz",
            "u_",
            "ua",
            "ub",
            "uc",
            "ud",
            "ue",
            "uf",
            "ug",
            "uh",
            "ui",
            "uj",
            "uk",
            "ul",
            "um",
            "un",
            "uo",
            "up",
            "uq",
            "ur",
            "us",
            "ut",
            "uu",
            "uv",
            "uw",
            "ux",
            "uy",
        ]
        self.set2_chars_14 = [
            "uz",
            "v_",
            "va",
            "vb",
            "vc",
            "vd",
            "ve",
            "vf",
            "vg",
            "vh",
            "vi",
            "vj",
            "vk",
            "vl",
            "vm",
            "vn",
            "vo",
            "vp",
            "vq",
            "vr",
            "vs",
            "vt",
            "vu",
            "vv",
            "vw",
            "vx",
            "vy",
            "vz",
            "w_",
            "wa",
            "wb",
            "wc",
            "wd",
            "we",
            "wf",
            "wg",
            "wh",
            "wi",
            "wj",
            "wk",
            "wl",
            "wm",
            "wn",
            "wo",
            "wp",
            "wq",
            "wr",
            "ws",
            "wt",
            "wu",
            "wv",
            "ww",
            "wx",
            "wy",
        ]
        self.set2_chars_15 = [
            "wz",
            "x_",
            "xa",
            "xb",
            "xc",
            "xd",
            "xe",
            "xf",
            "xg",
            "xh",
            "xi",
            "xj",
            "xk",
            "xl",
            "xm",
            "xn",
            "xo",
            "xp",
            "xq",
            "xr",
            "xs",
            "xt",
            "xu",
            "xv",
            "xw",
            "xx",
            "xy",
            "xz",
            "y_",
            "ya",
            "yb",
            "yc",
            "yd",
            "ye",
            "yf",
            "yg",
            "yh",
            "yi",
            "yj",
            "yk",
            "yl",
            "ym",
            "yn",
            "yo",
            "yp",
            "yq",
            "yr",
            "ys",
            "yt",
            "yu",
            "yv",
            "yw",
            "yx",
            "yy",
        ]
        self.set2_chars_16 = [
            "yz",
            "z_",
            "za",
            "zb",
            "zc",
            "zd",
            "ze",
            "zf",
            "zg",
            "zh",
            "zi",
            "zj",
            "zk",
            "zl",
            "zm",
            "zn",
            "zo",
            "zp",
            "zq",
            "zr",
            "zs",
            "zt",
            "zu",
            "zv",
            "zw",
            "zx",
            "zy",
            "zz",
        ]

        self.set3_chars = [
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "_ADJ_",
            "_ADP_",
            "_ADV_",
            "_CONJ_",
            "_DET_",
            "_NOUN_",
            "_NUM_",
            "_PRON_",
            "_PRT_",
            "_VERB_",
            "a",
            "b",
            "c",
            "d",
            "e",
            "f",
            "g",
            "h",
            "i",
            "j",
            "k",
            "l",
            "m",
            "n",
            "o",
            "other",
            "p",
            "punctuation",
            "q",
            "r",
            "s",
            "t",
            "u",
            "v",
            "w",
            "x",
            "y",
            "z",
        ]

        DATA_LOCATION_KEY = NGramIndexer.DEFAULT_DATA_LOCATION_KEY
        if (
            os.system(
                "cd "
                + os.path.join(
                    os.path.expanduser("~"),
                    self.ngram_indexer.configurations[
                        NGramIndexer.DEFAULT_DEFAULTNGRAM_KEY
                    ][DATA_LOCATION_KEY],
                )
            )
            == 0
        ):
            os.chdir(os.path.expanduser("~"))
            self.ngram_data_path = (
                "./"
                + self.ngram_indexer.configurations[
                    NGramIndexer.DEFAULT_DEFAULTNGRAM_KEY
                ][DATA_LOCATION_KEY]
            )
        else:
            os.chdir(os.path.expanduser("~"))
            self.ngram_data_path = (
                "./"
                + self.ngram_indexer.configurations[NGramIndexer.DEFAULT_USERNGRAM_KEY][
                    DATA_LOCATION_KEY
                ]
            )

    """1-grams-------------------------------------------------------------------------------------------------------"""

    def stage_1(self):
        responses = []
        for char in self.set1_chars:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-1gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-1gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 1)

    """2-grams-------------------------------------------------------------------------------------------------------"""

    def stage_2(self):
        responses = []
        for char in self.set2_chars_1:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-1gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 2)

    def stage_3(self):
        responses = []
        for char in self.set2_chars_2:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 3)

    def stage_4(self):
        responses = []
        for char in self.set2_chars_3:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 4)

    def stage_5(self):
        responses = []
        for char in self.set2_chars_4:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 5)

    def stage_6(self):
        responses = []
        for char in self.set2_chars_5:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 6)

    def stage_7(self):
        responses = []
        for char in self.set2_chars_6:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 7)

    def stage_8(self):
        responses = []
        for char in self.set2_chars_7:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 8)

    def stage_9(self):
        responses = []
        for char in self.set2_chars_8:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 9)

    def stage_10(self):
        responses = []
        for char in self.set2_chars_9:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 10)

    def stage_11(self):
        responses = []
        for char in self.set2_chars_10 + self.set2_chars_11:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 11)

    def stage_12(self):
        responses = []
        for char in self.set2_chars_12:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 12)

    def stage_13(self):
        responses = []
        for char in self.set2_chars_13:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 13)

    def stage_14(self):
        responses = []
        for char in self.set2_chars_14:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 14)

    def stage_15(self):
        responses = []
        for char in self.set2_chars_15:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 15)

    def stage_16(self):
        responses = []
        for char in self.set2_chars_16:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-2gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-2gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 16)

    """3-grams-------------------------------------------------------------------------------------------------------"""

    def stage_17(self):
        responses = []
        for char in self.set2_chars_1:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 17)

    def stage_18(self):
        responses = []
        for char in self.set2_chars_2:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 18)

    def stage_19(self):
        responses = []
        for char in self.set2_chars_3:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 19)

    def stage_20(self):
        responses = []
        for char in self.set2_chars_4:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 20)

    def stage_21(self):
        responses = []
        for char in self.set2_chars_5:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 21)

    def stage_22(self):
        responses = []
        for char in self.set2_chars_6:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 22)

    def stage_23(self):
        responses = []
        for char in self.set2_chars_7:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 23)

    def stage_24(self):
        responses = []
        for char in self.set2_chars_8:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 24)

    def stage_25(self):
        responses = []
        for char in self.set2_chars_9:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 25)

    def stage_26(self):
        responses = []
        for char in self.set2_chars_10 + self.set2_chars_11:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 26)

    def stage_27(self):
        responses = []
        for char in self.set2_chars_12:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 27)

    def stage_28(self):
        responses = []
        for char in self.set2_chars_13:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 28)

    def stage_29(self):
        responses = []
        for char in self.set2_chars_14:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 29)

    def stage_30(self):
        responses = []
        for char in self.set2_chars_15:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 30)

    def stage_31(self):
        responses = []
        for char in self.set2_chars_16:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-3gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-3gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 31)

    """4-grams-------------------------------------------------------------------------------------------------------"""

    def stage_32(self):
        responses = []
        for char in self.set2_chars_1:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 32)

    def stage_33(self):
        responses = []
        for char in self.set2_chars_2:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 33)

    def stage_34(self):
        responses = []
        for char in self.set2_chars_3:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 34)

    def stage_35(self):
        responses = []
        for char in self.set2_chars_4:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 35)

    def stage_36(self):
        responses = []
        for char in self.set2_chars_5:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 36)

    def stage_37(self):
        responses = []
        for char in self.set2_chars_6:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 37)

    def stage_38(self):
        responses = []
        for char in self.set2_chars_7:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 38)

    def stage_39(self):
        responses = []
        for char in self.set2_chars_8:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 39)

    def stage_40(self):
        responses = []
        for char in self.set2_chars_9:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 40)

    def stage_41(self):
        responses = []
        for char in self.set2_chars_10 + self.set2_chars_11:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 41)

    def stage_42(self):
        responses = []
        for char in self.set2_chars_12:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 42)

    def stage_43(self):
        responses = []
        for char in self.set2_chars_13:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 43)

    def stage_44(self):
        responses = []
        for char in self.set2_chars_14:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 44)

    def stage_45(self):
        responses = []
        for char in self.set2_chars_15:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 45)

    def stage_46(self):
        responses = []
        for char in self.set2_chars_16:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-4gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-4gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 46)

    """5-grams-------------------------------------------------------------------------------------------------------"""

    def stage_47(self):
        responses = []
        for char in self.set2_chars_1:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 47)

    def stage_48(self):
        responses = []
        for char in self.set2_chars_2:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 48)

    def stage_49(self):
        responses = []
        for char in self.set2_chars_3:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 49)

    def stage_50(self):
        responses = []
        for char in self.set2_chars_4:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 50)

    def stage_51(self):
        responses = []
        for char in self.set2_chars_5:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 51)

    def stage_52(self):
        responses = []
        for char in self.set2_chars_6:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 52)

    def stage_53(self):
        responses = []
        for char in self.set2_chars_7:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 53)

    def stage_54(self):
        responses = []
        for char in self.set2_chars_8:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 54)

    def stage_55(self):
        responses = []
        for char in self.set2_chars_9:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 55)

    def stage_56(self):
        responses = []
        for char in self.set2_chars_11:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 56)

    def stage_57(self):
        responses = []
        for char in self.set2_chars_12:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 57)

    def stage_58(self):
        responses = []
        for char in self.set2_chars_13:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 58)

    def stage_59(self):
        responses = []
        for char in self.set2_chars_14:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 59)

    def stage_60(self):
        responses = []
        for char in self.set2_chars_15:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 60)

    def stage_61(self):
        responses = []
        for char in self.set2_chars_16:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-5gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-5gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 61)

    """dependencies--------------------------------------------------------------------------------------------------"""

    def stage_62(self):
        responses = []
        for char in self.set3_chars:
            file_url = (
                "https://storage.googleapis.com/books/ngrams/books/googlebooks-eng-all-0gram-20120701-"
                + char
                + ".gz"
            )
            file_name = "googlebooks-eng-all-0gram-20120701-" + char
            file(file_url, self.ngram_data_path)
            decompress_gz(self.ngram_data_path + "/" + file_name + ".gz")
            response_tuple = self.ngram_indexer.index_ngram_file_bulk(
                file_name, "stats"
            )
            response = list(response_tuple[1])
            self.monitoring_file(response, file_name)
            responses = responses + response

        self.monitoring_stage(responses, 62)

    def execute(self):
        time.sleep(2)
        self.stage_1()
        time.sleep(2)
        self.stage_2()
        time.sleep(2)
        self.stage_3()
        time.sleep(2)
        self.stage_4()
        time.sleep(2)
        self.stage_5()
        time.sleep(2)
        self.stage_6()
        time.sleep(2)
        self.stage_7()
        time.sleep(2)
        self.stage_8()
        time.sleep(2)
        self.stage_9()
        time.sleep(2)
        self.stage_10()
        time.sleep(2)
        self.stage_11()
        time.sleep(2)
        self.stage_12()
        time.sleep(2)
        self.stage_13()
        time.sleep(2)
        self.stage_14()
        time.sleep(2)
        self.stage_15()
        time.sleep(2)
        self.stage_16()
        time.sleep(2)
        self.stage_17()
        time.sleep(2)
        self.stage_18()
        time.sleep(2)
        self.stage_19()
        time.sleep(2)
        self.stage_20()
        time.sleep(2)
        self.stage_21()
        time.sleep(2)
        self.stage_22()
        time.sleep(2)
        self.stage_23()
        time.sleep(2)
        self.stage_24()
        time.sleep(2)
        self.stage_25()
        time.sleep(2)
        self.stage_26()
        time.sleep(2)
        self.stage_27()
        time.sleep(2)
        self.stage_28()
        time.sleep(2)
        self.stage_29()
        time.sleep(2)
        self.stage_30()
        time.sleep(2)
        self.stage_31()
        time.sleep(2)
        self.stage_32()
        time.sleep(2)
        self.stage_33()
        time.sleep(2)
        self.stage_34()
        time.sleep(2)
        self.stage_35()
        time.sleep(2)
        self.stage_36()
        time.sleep(2)
        self.stage_37()
        time.sleep(2)
        self.stage_38()
        time.sleep(2)
        self.stage_39()
        time.sleep(2)
        self.stage_40()
        time.sleep(2)
        self.stage_41()
        time.sleep(2)
        self.stage_42()
        time.sleep(2)
        self.stage_43()
        time.sleep(2)
        self.stage_44()
        time.sleep(2)
        self.stage_45()
        time.sleep(2)
        self.stage_46()
        time.sleep(2)
        self.stage_47()
        time.sleep(2)
        self.stage_48()
        time.sleep(2)
        self.stage_49()
        time.sleep(2)
        self.stage_50()
        time.sleep(2)
        self.stage_51()
        time.sleep(2)
        self.stage_52()
        time.sleep(2)
        self.stage_53()
        time.sleep(2)
        self.stage_54()
        time.sleep(2)
        self.stage_55()
        time.sleep(2)
        self.stage_56()
        time.sleep(2)
        self.stage_57()
        time.sleep(2)
        self.stage_58()
        time.sleep(2)
        self.stage_59()
        time.sleep(2)
        self.stage_60()
        time.sleep(2)
        self.stage_61()
        time.sleep(2)
        self.stage_62()
        time.sleep(2)

    def monitoring_file(self, response, file_name):
        if len(response) == 0:
            return
        else:
            content = (
                "Something is wrong with file "
                + file_name
                + ", file gave following responses\n\n"
                + str(response)
            )
            mail = smtplib.SMTP("smtp.gmail.com", 587)
            mail.ehlo()
            mail.starttls()
            mail.login("juan@mammut.io", "13692468$/")
            mail.sendmail("juan@mammut.io", "juan@mammut.io", content)
            mail.close()

    def monitoring_stage(self, responses, stage_num):
        if len(responses) == 0:
            content = (
                "Everything OK with stage "
                + str(stage_num)
                + "next stages will start in seconds!"
            )
            mail = smtplib.SMTP("smtp.gmail.com", 587)
            mail.ehlo()
            mail.starttls()
            mail.login("juan@mammut.io", "13692468$/")
            mail.sendmail("juan@mammut.io", "juan@mammut.io", content)
            mail.close()
        else:
            content = (
                "Something went wrong with "
                + stage_num
                + " file reports should have arrived already, "
                "please check your inbox"
            )
            mail = smtplib.SMTP("smtp.gmail.com", 587)
            mail.ehlo()
            mail.starttls()
            mail.login("juan@mammut.io", "13692468$/")
            mail.sendmail("juan@mammut.io", "juan@mammut.io", content)
            mail.close()



from mammut.assessment.problem import Problem
from mammut.curriculum.models.hugging_face.roberta import (
    HuggingFaceRobertaTokenClassificationModel,
)
from typing import Dict
import torch


class RobertaTokenClassificationProblemMFT(Problem):
    """This class evaluates the encoding produced by a token classification
    model.

    Test the minimal functionality: expected token(s) is(are) found in the
    encoded output.
    """

    def __init__(
        self,
        course_id: str,
        assessment_id: str,
        schedule: str,
        problem_id: str,
        problem_type: str,
        parameters: str,
        language_competence: str,
        weight: str,
        retry: str,
        observations: str,
    ):
        """
        Args:
            course_id(str): id of the course to which this problem belongs.
            assessment_id(str): id of the assessment that bundles this problem.
            schedule(str): lesson schedule to execute this problem.
            problem_id(str): id for this problem.
            problem_type(str): problem type enumeration string.
            parameters (str): JSON string parameters for this problem.
            language_competence(str): language competence enumeration value for this problem.
            weight(str): float string value for the weight of this particular problem
                in the overall course score.
            observations(str): observations string. Only descriptive.
        """
        Problem.__init__(
            self,
            course_id,
            assessment_id,
            schedule,
            problem_id,
            problem_type,
            parameters,
            language_competence,
            weight,
            retry,
            observations,
        )
        self._input: str = self.parameters["input"]
        self._expected_classification_dict = self.parameters[
            "expected_classification"
        ]
        self._classification_output: str = ""

    def _token_classification(self, examples, model):

        tokens = model._tokenizer.tokenize(
            model._tokenizer.decode(model._tokenizer.encode(examples))
        )
        inputs = model._tokenizer.encode(
            examples, return_tensors="pt", padding=True, truncation=False
        )
        outputs = model(inputs).logits
        predictions = torch.argmax(outputs, dim=2)
        prediction = predictions[0].tolist()
        classification = [
            (token, model.labels[pred])
            for token, pred in zip(tokens, prediction)
        ]
        results = dict((y, x) for x, y in classification)
        return results

    def solve(
        self, model: HuggingFaceRobertaTokenClassificationModel, **kwargs
    ) -> bool:
        self._retries_count += 1
        self._classification_output = self._token_classification(
            self._input, model
        )
        if self._classification_output == self._expected_classification_dict:
            return True
        else:
            return False

    def get_result_info(self) -> Dict:
        info = {
            "Tokens classification result": ", ".join(
                self._classification_output
            ),
            "Expected classification": self._expected_classification_dict,
        }
        return info



import enum


class Phase:
    """Constants for typed phase description in messages.
    """

    LOAD_PACKAGE_DATA = "Load Package Data"
    LOAD_ASSESSMENT_DATA = "Load Assessment Data"
    LOAD_CURRICULUM_LESSONS_DATA = "Load Lessons Data"
    CREATE_CURRICULUM_FLOWCHART = "Create Curriculum Flowchart"
    CREATE_CLASSROOMS = "Create Classrooms"
    LECTURE_LESSON = "Lecture Lesson"


class Component(enum.Enum):
    """Component responsible for the message report.
    """

    Curriculum = 1
    Classroom = 2
    Course = 3
    Lesson = 4
    Problem = 5
    Package = 6
    LinguisticStandard = 7



# coding=utf-8

from abc import *
import tensorflow as tf
import time
import os.path

from tensorflow.python.platform import tf_logging as logging
from mammut.models.trainable_node import TrainableNode
from mammut.models import DISTRIBUTED_TRAINING
from mammut.models.parameter_node import PARAMETER_NODE_JOB_NAME
from mammut.models.configurable import Configurable


class Trainer(Configurable):
    def __init__(
        self, configuration_path, configuration_name, redis_connection_pool=None
    ):
        Configurable.__init__(
            self, configuration_path, configuration_name, redis_connection_pool
        )

    @abstractmethod
    def _initialize_graph(self):
        ...

    # TODO: Este metodo no se llamara hasta setear el init_fn del supervisor.
    @abstractmethod
    def _recover_variables(self, sess):
        ...

    @abstractmethod
    def _build_model(self) -> TrainableNode:
        ...

    @abstractmethod
    def _get_input_placeholder(self):
        ...

    @abstractmethod
    def _get_expectedResult_placeholder(self):
        ...

    @abstractmethod
    def _fill_feed_dict(self, sess, input_placeholder, expected_result_placeholder):
        ...

    @abstractmethod
    def _get_current_epoch(self):
        ...

    def train(
        self,
        logdir,
        max_epochs,
        ps_hosts=None,
        worker_hosts=None,
        job_name=None,
        task_index=None,
        save_summaries_secs=200,
        save_model_secs=300,
        log_step_mod=100,
    ):

        graph = tf.Graph()
        with graph.as_default():

            is_chief = True  ## Default value
            target = ""
            is_parameter_server = False

            if DISTRIBUTED_TRAINING:
                logging.info("Distributed training active")
                if ps_hosts != None:
                    # Create a cluster from the parameter server and worker hosts.
                    cluster = tf.train.ClusterSpec(
                        {"ps": ps_hosts, "worker": worker_hosts}
                    )
                    # Create and start a server for the local task.
                    server = tf.train.Server(
                        cluster, job_name=job_name, task_index=task_index
                    )
                    if job_name == PARAMETER_NODE_JOB_NAME:
                        is_parameter_server = True
                        server.join()
                    # Check if the supervisor must be mark as chief
                    is_chief = task_index == 0
                    target = server.target

            if not is_parameter_server:
                if DISTRIBUTED_TRAINING:
                    with tf.device(
                        "/job:{}/task:{:d}".format(PARAMETER_NODE_JOB_NAME, 0)
                    ):
                        # step counter used in summaries and checkpoint filenames.
                        global_step = tf.Variable(
                            0, trainable=False, name="global_step"
                        )
                else:
                    global_step = tf.Variable(0, trainable=False, name="global_step")

                # Build the model
                self._initialize_graph()
                trainable_node = self._build_model()

                # Get the training step
                input_placeholder = self._get_input_placeholder()
                expected_result_placeholder = self._get_expectedResult_placeholder()

                # operations
                training_step_op = trainable_node.training_step_op(
                    global_step, input_placeholder, expected_result_placeholder
                )

                # Build the summary Tensor based on the TF collection of Summaries.
                summary_op = tf.summary.merge_all()
                training_step_op = [summary_op] + training_step_op

                # Add ops to save and restore all the variables.
                # Saver only captures all variables at the time it is constructed, so any new variables defined
                # after Saver() will not be saved/restored automatically.
                saver = None
                summary_writer = None
                if logdir != None:
                    saver = tf.train.Saver()
                    summary_writer = tf.summary.FileWriter(logdir, graph)

                # TODO: This class tf.train.Supervisor is deprecated. Please use tf.train.MonitoredTrainingSession instead.
                # Create a "supervisor", which oversees the training process.
                sv = tf.train.Supervisor(
                    is_chief=is_chief,
                    logdir=logdir,
                    # init_op=None, # Init function to manually recover a model
                    summary_op=None,
                    saver=saver,
                    global_step=global_step,
                    save_summaries_secs=save_summaries_secs,
                    save_model_secs=save_model_secs,
                    summary_writer=summary_writer,
                    init_fn=self._recover_variables,
                )

                # TODO: Agregar parametro para configurar el uso de init_op/init_fn

                with sv.managed_session(
                    master=target,
                    config=tf.ConfigProto(
                        allow_soft_placement=True, log_device_placement=True
                    ),
                ) as sess:

                    while (
                        not sv.should_stop() and self._get_current_epoch() < max_epochs
                    ):
                        start_time = time.time()

                        # Feed in data for session run
                        feed_dict = self._fill_feed_dict(
                            sess, input_placeholder, expected_result_placeholder
                        )

                        # Run session
                        summary, _, loss_value, step = sess.run(
                            training_step_op, feed_dict=feed_dict
                        )

                        # Append summary
                        sv.summary_computed(sess, summary)

                        current_step = int(step)

                        if current_step % log_step_mod == 0:
                            duration = time.time() - start_time
                            logging.info(
                                "Step %d: loss = %.2f (%.3f sec)"
                                % (step, loss_value, duration)
                            )

                sv.stop()

                return float(loss_value)

            else:
                return -1.0

    # TODO: falta colocar la logica para automaticamente evaluar el modelo contra los datasets de validacion y pruebas cada cierta cantidad de epocas



from mammut import BASE_MODEL_VISON_DIR
from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.constants import CorpusTable
from mammut.common.corpus.corpus_base import Corpus, CorpusScenarioEntry
from mammut.common.corpus.crawler_corpus import CrawlerCorpus
from mammut.common.google_api import GoogleStorageDrive
from mammut.common.lexicon.dictionary import Dictionary
from mammut.common.storage.storage_manager import StorageManager


class ImagesCorpus(Corpus):
    DEFAULT_SHEET_TITTLE = "images_corpus"
    DEFAULT_BEGIN_RANGE = "A2"
    DEFAULT_IMAGE_ID = 1

    DEFAULT_SCENARIO_TYPE_VALUE = "Monologue"
    DEFAULT_EVENT_MESSAGE_ORIGINAL_VALUE = ""
    DEFAULT_EVENT_MESSAGE_VALUE = '[variable_image|"%s"]'
    DEFAULT_HIDDEN_VALUE = ""
    DEFAULT_FIELD_VALUE = ""
    DEFAULT_SOURCE_VALUE = "Mammut"
    DEFAULT_COMPLEXITY_VALUE = ""
    storage_manager = StorageManager()
    # configurations keys
    FOLDER_ID = "folder_id"  # del drive
    DESTINATION_DIRECTORY = "destination_directory"

    def __init__(
        self,
        spreadsheet_id: str,
        id,
        corpus_sheet_title: str,
        corpus_extension_sheet_title: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        regional_settings: str,
        update_status_text_box,
        annotable: bool,
        configuration_json,
    ):
        self.sheet_title = corpus_sheet_title
        self.spreadsheet_id = spreadsheet_id
        self.destination_folder = (
            BASE_MODEL_VISON_DIR + "/train/vision/" + spreadsheet_id + "/"
        )
        Corpus.__init__(
            self,
            spreadsheet_id,
            id,
            corpus_sheet_title,
            corpus_extension_sheet_title,
            tokenizer,
            dictionary,
            regional_settings,
            annotable,
            configuration_json,
        )

    def create_dump_sheet(self, spreadsheet_id):
        temporal_sheet_title = self.storage_manager.create_temporal_new_sheet(
            spreadsheet_id, self.sheet_title
        )
        self.storage_manager.add_rows_to_spreadsheet(
            spreadsheet_id,
            temporal_sheet_title,
            CrawlerCorpus.DEFAULT_COLUMN_TITTLES,
            CrawlerCorpus.DEFAULT_START_ROW_TITTLE,
            CrawlerCorpus.DEFAULT_START_COLUMN_TITTLE,
        )
        return temporal_sheet_title

    def get_images_data(self):
        folders_tuple = GoogleStorageDrive.get_folders_in_folder(
            self.configuration[ImagesCorpus.FOLDER_ID]
        )
        all_images = list()
        for folder in folders_tuple:
            all_images += self.download_images(folder[0], folder[1])
        return all_images

    def download_images(self, folder_name, folder_id):
        images = GoogleStorageDrive.get_all_files_in_folder(folder_id)
        images_titles = [i[0] for i in images]
        GoogleStorageDrive.download_files(
            folder_id,
            self.destination_folder + "/" + folder_name,
            images_titles,
        )
        return images_titles

    def dump_corpus(self, images):
        data = []
        for index, image in enumerate(images):
            image_name = ".".join(image.split(".")[:-1])
            data.append(
                [
                    ImagesCorpus.DEFAULT_IMAGE_ID,
                    index,
                    ImagesCorpus.DEFAULT_SCENARIO_TYPE_VALUE,
                    ImagesCorpus.DEFAULT_EVENT_MESSAGE_ORIGINAL_VALUE,
                    ImagesCorpus.DEFAULT_EVENT_MESSAGE_VALUE % image_name,
                    ImagesCorpus.DEFAULT_HIDDEN_VALUE,
                    ImagesCorpus.DEFAULT_FIELD_VALUE,
                    ImagesCorpus.DEFAULT_SOURCE_VALUE,
                    self.regional_settings,
                    ImagesCorpus.DEFAULT_COMPLEXITY_VALUE,
                ]
            )
        self.write_data(self.DEFAULT_CORPUS_RANGE, data)

    def write_data(self, range, rows):
        sheet_name, range_cell = range.split("!")
        self.storage_manager.add_rows_to_spreadsheet(
            self.spreadsheet_id,
            sheet_name,
            rows,
            range_cell[1 : len(range_cell)],
            range_cell[0],
        )

    def load_data(self):
        images_titles = self.get_images_data()
        # TODO:el nombre que se guarda en el coprus del exel debe contenr el nombre del folder donde se guarda?
        self.dump_corpus(images_titles)

    def retrain_model(self):
        return

    def get_corpus_scenario_entry_from_row(self, g):
        return CorpusScenarioEntry(
            g[0],
            g[1].iloc[0][CorpusTable.SCENARIO_TYPE_COLUMN_NAME()],
            g[1],
            self.tokenizer,
            self.dictionary,
            self.regional_settings,
            self,
        )



# coding=utf-8

__all__ = [
    "annotations_validator_widget",
    "annotations_widget",
    "disambiguation_widget",
    "image_retraining_widget",
    "lexicalization_widget",
    "room_form",
    "simulator_widget",
    "simulator_widget_compilation_utils",
    "simulator_widget_operation_utils",
    "simulator_widget_simulation_utils",
    "solve_thought_offer_to_be_learn",
    "user_form",
    "widgets_utilities",
]



from mammut.common.storage.storage_manager import StorageManager
from mammut.common.catalog.hugging_face_datasets import TextOnly


class Catalog:
    """Catalog class allows find out interactively
     which documents do we have that are valid and
     meaningful.

     Attributes:
        catalog_data: Catalog class has a dataframe
            with the information of all internal documents
            (spreadsheet or presentations) that can used
            in different processes.
     """

    storage_manager = StorageManager()

    FEATURE_NAME_COLUMN = "source_name"
    FEATURE_TYPE_COLUMN = "source_type"
    FEATURE_SUBTYPE_COLUMN = "source_subtype"
    FEATURE_ID_COLUMN = "source_id"
    FEATURE_DESCRIPTION_COLUMN = "source_description"
    FEATURE_REGIONAL_SETTINGS_COLUMN = "regional_settings"
    FEATURE_DEFAULT_COLUMN = "default"

    def __init__(self):
        self.catalog_data = self.storage_manager.get_spreadsheet_as_dataframe(
            "1xwP87zj5m71-MtzTeAuwbGP2-FgPaM5imbVcCMcJudU",
            "catalog",
            "A",
            "1",
            "G",
            False,
        )

    def show_catalog_data(self):
        """Returns the complete catalog.
        """
        return self.catalog_data

    def show_info_source_type(self, source_type):
        """returns the catalog filtered by source_type.

        Arg:
            source_type(str): source type, it can be
            standard, package, documentation or
            annotated_corpus.
        """
        for r, c in self.catalog_data.iterrows():
            if c[Catalog.FEATURE_TYPE_COLUMN] == source_type:
                return (
                    c[Catalog.FEATURE_NAME_COLUMN],
                    c[Catalog.FEATURE_ID_COLUMN],
                )

    def get_default_id(self, source_type):
        """returns a default id of a standard document.

        Args:
            source_type(str): source type, it can be
            standard, package, documentation or
            annotated_corpus.
        """
        for r, c in self.catalog_data.iterrows():
            if c[Catalog.FEATURE_TYPE_COLUMN] == source_type:
                if c[Catalog.FEATURE_DEFAULT_COLUMN] == "TRUE":
                    return c[Catalog.FEATURE_ID_COLUMN]

    def get_standard_id_subtype(self, source_subtype, regional_settings):
        """Returns the id of some model developed by
        Mammut in English or Spanish.

        Args:
            source_subtype(str): standard document subtype,
            it can be conjunctions, com_conjunctions,
            prepositions, com_prepositions
            articles, deictics.

            regional_settings(str): model language, en or es.
        """
        for r, c in self.catalog_data.iterrows():
            if c[Catalog.FEATURE_TYPE_COLUMN] == "standard":
                if c[Catalog.FEATURE_SUBTYPE_COLUMN] == source_subtype:
                    if (
                        c[Catalog.FEATURE_REGIONAL_SETTINGS_COLUMN]
                        == regional_settings
                    ):
                        return c[Catalog.FEATURE_ID_COLUMN]

    def get_external_corpus(self, corpus_type):
        """Returns a corpus sentences made with a huggingface dataset.

        Arg:
            corpus_type(str): for now, this implementation only
            supports single feature dataset.
        """
        if corpus_type == "HF_TO":
            dataset = TextOnly()
            corpus = dataset.data_translator()
            return corpus



# coding=utf-8
from elasticsearch import Elasticsearch
from mammut import ES_HOSTS, CONCORDANCE_QUERY_SIZE
from mammut.common import corpus
from mammut.common.corpus.constants import CorpusTable


class Concordance:
    NUMBER_OF_WORDS = 8
    NUMBER_OF_CHARS = 50
    SEPARATOR_CHARS = ["*", "(", ")"]

    def __init__(self):
        self.elasticsearch = Elasticsearch(ES_HOSTS)

    def semantic_split(self, sentence, word):
        for i in range(0, len(sentence)):
            if sentence[i] == word[0] and (i - 1 == -1 or sentence[i - 1] == " "):
                if sentence[i:][: len(word)] == word:
                    bottom_sentence = sentence[:i]
                    new_sentence = sentence[i + len(word) :]
                    if len(new_sentence) != 0:
                        if new_sentence[0] == " " or not new_sentence[0].isalpha():
                            return [bottom_sentence] + self.semantic_split(
                                new_sentence, word
                            )
                    else:
                        return [bottom_sentence] + [""]
        return [sentence]

    def semantic_contains(self, sentence, word):
        for i in range(0, len(sentence)):
            if sentence[i] == word[0] and (i - 1 == -1 or sentence[i - 1] == " "):
                if sentence[i:][: len(word)] == word:
                    front_sentence = sentence[i + len(word) :]
                    if len(front_sentence) != 0:
                        if front_sentence[0] == " " or not front_sentence[0].isalpha():
                            return True
                    else:
                        return True
        return False

    def get_data_from_index(self, index, query):
        limits = query.split(Concordance.SEPARATOR_CHARS[1])
        by_regex = True
        if len(limits) == 1:
            by_regex = False
        if by_regex:
            bottom_limit = limits[0]
            front_limit_and_regex = limits[1].split(Concordance.SEPARATOR_CHARS[2])
            front_limit = front_limit_and_regex[1]
            regex = front_limit_and_regex[0]
            tonkens_number_str = regex.split(Concordance.SEPARATOR_CHARS[0])[1]
            if len(tonkens_number_str) == 0:
                tonkens_number = -1
            else:
                tonkens_number = int(tonkens_number_str)
            query_result = self.elasticsearch.search(
                index=index,
                body={
                    "query": {
                        "dis_max": {
                            "queries": [
                                {
                                    "bool": {
                                        "must": [
                                            {
                                                "match_phrase": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): front_limit
                                                }
                                            },
                                        ],
                                        "should": [
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + " "
                                                }
                                            },
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + ","
                                                }
                                            },
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + "."
                                                }
                                            },
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + ":"
                                                }
                                            },
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + ";"
                                                }
                                            },
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + "?"
                                                }
                                            },
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + "!"
                                                }
                                            },
                                            {
                                                "match_phrase_prefix": {
                                                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): bottom_limit
                                                    + "\n"
                                                }
                                            },
                                        ],
                                    }
                                }
                            ]
                        }
                    },
                    "size": CONCORDANCE_QUERY_SIZE,
                },
            )
            results = []
            for i in query_result["hits"]["hits"]:
                remainders = self.semantic_split(
                    i["_source"][CorpusTable.EVENT_MESSAGE_COLUMN_NAME()].lower(),
                    bottom_limit.lower(),
                )
                for j in range(0, len(remainders)):
                    result = ""
                    if j + 1 == len(remainders):
                        break
                    else:
                        if remainders[j + 1] != "":
                            if not self.semantic_contains(
                                remainders[j + 1], front_limit.lower()
                            ):
                                pass
                            else:
                                middle_words_and_front_words = self.semantic_split(
                                    remainders[j + 1], front_limit.lower()
                                )
                                middle_words = middle_words_and_front_words[0]
                                if (
                                    len(middle_words.split(" ")) == tonkens_number + 2
                                    or tonkens_number == -1
                                ):
                                    words_behind = remainders[j].split(" ")
                                    words_front = middle_words_and_front_words[1].split(
                                        " "
                                    )
                                    bottom = (
                                        len(words_behind) - Concordance.NUMBER_OF_WORDS
                                    )
                                    if bottom < 0:
                                        bottom = 0
                                    bottom_text = " ".join(words_behind[bottom:])
                                    if len(bottom_text) < Concordance.NUMBER_OF_CHARS:
                                        bottom_text += "_" * (
                                            Concordance.NUMBER_OF_CHARS
                                            - len(bottom_text)
                                        )
                                    result = (
                                        bottom_text
                                        + " <b>"
                                        + bottom_limit.lower()
                                        + " "
                                        + middle_words
                                        + " "
                                        + front_limit
                                        + "</b> "
                                        + " ".join(
                                            words_front[: Concordance.NUMBER_OF_WORDS]
                                        )
                                    )

                    if result != "":
                        results.append(result)
            return results
        else:
            query_result = self.elasticsearch.search(
                index=index,
                body={
                    "query": {
                        "dis_max": {
                            "queries": [
                                {
                                    "match_phrase": {
                                        CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): query
                                    }
                                }
                            ]
                        }
                    },
                    "size": CONCORDANCE_QUERY_SIZE,
                },
            )
            results = []
            for i in query_result["hits"]["hits"]:
                remainders = self.semantic_split(
                    i["_source"][CorpusTable.EVENT_MESSAGE_COLUMN_NAME()].lower(),
                    query.lower(),
                )
                for j in range(0, len(remainders)):
                    result = ""
                    if j + 1 == len(remainders) or remainders[j + 1] == "":
                        break
                    else:
                        if remainders[j + 1] != "":
                            words_behind = remainders[j].split(" ")
                            words_front = remainders[j + 1].split(" ")
                            bottom = len(words_behind) - Concordance.NUMBER_OF_WORDS
                            if bottom < 0:
                                bottom = 0
                            bottom_text = " ".join(words_behind[bottom:])
                            if len(bottom_text) < Concordance.NUMBER_OF_CHARS:
                                bottom_text += "_" * (
                                    Concordance.NUMBER_OF_CHARS - len(bottom_text)
                                )
                            result = (
                                bottom_text
                                + " <b>"
                                + query
                                + "</b> "
                                + " ".join(words_front[: Concordance.NUMBER_OF_WORDS])
                            )

                    if result != "":
                        results.append(result)
            return results
